import asyncio
import socket
import ipaddress
import pprint
from .net import *

# TODO: address doesn't support domain resolution
# from a specific interface. This may not matter though.
class Address():
    def __init__(self, host, port, route=None, sock_type=socket.SOCK_STREAM, timeout=2):
        self.timeout = timeout
        self.resolved = False
        self.sock_type = sock_type
        self.route = route
        self.host = host
        self.host_type = None
        self.port = int(port)
        self.afs_found = [] # Supported AFs once resolved.
        self.ips = {} # Destination indexed by IP.
        self.host = to_s(host) if host is not None else host
        log("> Address: %s:%d" % (self.host, self.port))

    async def res(self):
        # Lookup IPs for domain.
        route = self.route
        loop = asyncio.get_event_loop()

        # Determine if IP or domain.
        self.chosen = AF_ANY
        if self.host is None:
            return

        # Try parse host as an IP address.
        self.af = route.af
        try:
            self.ip_obj = ipaddress.ip_address(self.host)
            if self.ip_obj.version == 4:
                if self.af not in [AF_ANY, IP4]:
                    raise Exception("Found IP doesn't match AF.")

                self.chosen = socket.AF_INET
                self.ips[AF_ANY] = self.ips[socket.AF_INET] = self.host
            else:
                if self.af not in [AF_ANY, IP6]:
                    raise Exception("Found IP doesn't match AF.")

                self.chosen = socket.AF_INET6
                self.ips[AF_ANY] = self.ips[socket.AF_INET6] = self.host

            self.host_type = HOST_TYPE_IP
        except Exception as e:
            self.ip_obj = None
            self.host_type = HOST_TYPE_DOMAIN
            self.chosen = self.af

        # Set IP port tup.
        self.as_tup = self.tup = self.tuple = ()

        # Set target to resolve.
        if self.host_type == HOST_TYPE_IP:
            # Target is an IP.
            target = self.ips[self.chosen]

            # Patch link local addresses.
            if self.af == IP6 and target not in ["::", "::1"]:
                target = ip6_patch_bind_ip(
                    self.ip_obj,
                    target,
                    route.interface
                )
        else:
            # Target is a domain name.
            target = self.host

        # Get endpoint connect / bind results.
        results = []
        afs_wanted = VALID_AFS if self.chosen == AF_ANY else [self.chosen]
        family = 0 if len(afs_wanted) == 2 else afs_wanted[0]
        try:
            addr_infos = loop.getaddrinfo(
                target,
                self.port,
                type=self.sock_type,
                family=family
            )
            results = await asyncio.wait_for(addr_infos, self.timeout)
        except Exception as e:
            log(f"{target} {self.port} {family} {self.sock_type}")
            log_exception()
            log("> Address: res e = %s" % (str(e)))

        # Choose a tuple that matches requirements.
        addr_tup = None
        afs_found = []  
        for addr_info in results:
            for af in afs_wanted:
                if af == addr_info[0]:
                    addr_tup = addr_info[4]
                    if af not in afs_found:
                        afs_found.append(af)

        # Strip %n from tup 0 if it exists.
        """
        addr_tup = (
            ip_strip_if(addr_tup[0]),
        ) + addr_tup[1:]
        """

        # No results found for our requirements.
        if addr_tup is None:
            log("> Address: res couldnt find compatible address")
            self.chosen = AF_NONE
            raise Exception("couldnt translate address")
        else:
            self.tup = addr_tup

        # Set attributes of the IP like if private or loopback.
        self.ip_set_info(ip_strip_if(addr_tup[0]))
        self.afs_found = afs_found
        self.resolved = True
        return self

    def __await__(self):
        return self.res().__await__()

    def supported(self):
        return self.afs_found

    def ip_set_info(self, ip_s):
        # Determine whether IP is public or private.
        ip_obj = ip_f(ip_s)
        if ip_obj.is_private:
            self.is_private = True
            self.is_public = False
        else:
            self.is_private = False
            self.is_public = True

        # Used for IPv6 code.
        if str(ip_obj) in VALID_LOOPBACKS:
            self.is_loopback = True
        else:
            self.is_loopback = False

    def target(self):
        return self.tup[0]

    def as_tuple(self):
        return self.as_tup

    def to_dict(self):
        return {
            "host_type": self.host_type,
            "host": self.host,
            "port": self.port,
            "sock_type": self.sock_type,
            "timeout": self.timeout,
            "ips": self.ips,
            "chosen": self.chosen,
            "afs_found": self.afs_found,
            "is_private": self.is_private,
            "is_public": self.is_public,
            "is_loopback": self.is_loopback,
            "tup": self.tup,
            "resolved": self.resolved
        }

    @staticmethod
    def from_dict(d):
        a = Address(
            host=d["host"],
            port=d["port"],
            sock_type=d["sock_type"],
            timeout=d["timeout"]
        )

        for key in d:
            setattr(a, key, d[key])
            #a.__setattr__("self." + key, d[key])

        return a

    # Pickle.
    def __getstate__(self):
        return self.to_dict()

    # Unpickle.
    def __setstate__(self, state):
        o = self.from_dict(state)
        self.__dict__ = o.__dict__

    # Show a representation of this object.
    def __repr__(self):
        return f"Address.from_dict({str(self)})"

    # Make this interface printable because it's useful.
    def __str__(self):
        return pprint.pformat(self.tup)

    def __hash__(self):
        return hash(repr(self))

    def  __len__(self):
        return 0

async def test_address(): # pragma: no cover
    from p2pd.interface import init_p2pd, Interface
    i = await Interface()
    print(i)
    a = await Address("www.google.com", 80, i.route()).res()

    d = a.to_dict()
    print(d)

    x = Address.from_dict(d)
    print(str(x))

    y = repr(x)
    print(type(y))

if __name__ == "__main__": # pragma: no cover
    from .interface import Interface
    from .net import IP6
    from .utils import timestamp, async_test
    async_test(test_address)



"""
Taken directly from here: https://github.com/talkiq/pystun3
I'm forking it because I want to add small changes

changes:
- full async support
- add more stun servers
- test that they all work
- improve error checking
- add ipv6 support
- improve commenting
- fix some smol bugs (including nat test bugs)
- load balancing to avoid overloading
- result average support to avoid invalid servers
- separate list of hosts that support ipv6 for less failures on ipv6
- TCP support (for get mappings)
- delta n test (some nats have predictable mappings and assign then a delta apart)
- added better checking for 'change IPs' and made a new address family for hosts that return correct change ip responses. the nat determination code needs to use these hosts. for regular 'get wan ip' and 'get port mapped' lookups you can use the change hosts or the mapping hosts (larger list)

Note 1: Some of the response times for DNS lookups to the STUN servers in
this module are on the order of 1 second or higher -- an astronomical
amount of time for a network. I have tried to use concurrency patterns
where ever possible to avoid delaying other, faster lookups.4

Note 2: I've read the STUN RFC and it seems to indicate that many of the fields in the protocol format take place over byte boundaries. Yet the client code here works on all the servers I've tested it on and doesn't make these assumptions. It's possible the spec is wrong or maybe my code just won't work with particular features of the STUN protocol. No idea.

TODO: sort the hosts by how fast they respond to a STUN request from domain resolution to reply time.
TODO: It seems that this is a pattern that reoccurs in several functions.
The general form might also make sense to add to the Net module.
TODO: Refactor code. The code in this module offers many good features but the code reflects too much cruft. It could do with a good cleanup.
"""

import time
import binascii
import random
import socket
import asyncio
import ipaddress
import copy
import struct
import re
from .utils import *
from .net import *
from .address import *
from .nat import *
from .base_stream import pipe_open
from .settings import *

"""
In most requests the max STUN time is given by:
    (
        (
            (addr_retry * (dns_timeout + con_timeout_if_applicable))
                + 
            (packet_retry * recv_timeout)
        ) * retry_no
    ) * consensus[1]

Put another way:
    for consensus [1]:
        for retry_no:
            for addr_retry:
                dns_timeout
                optional: con_timeout

            for packet_retry:
                recv_timeout

I don't know if my formula is correct though.
"""

STUN_CONF = dict_child({
    # Retry N times on reply timeout.
    "packet_retry": 2,

    # Retry N times on invalid address.
    "addr_retry": 2,

    # Seconds to use for a DNS request before timeout exception.
    "dns_timeout": 2,

    # N of M required for agreement.
    "consensus": [1, 1],

    # Retry no -- if an individual call fails how
    # many times to retry the whole thing.
    "retry_no": 3,

    # Reuse address tuple for bind() socket call.
    "reuse_addr": True,
}, NET_CONF)

# stun consts
MappedAddress = '0001'
SourceAddress = '0004'
ChangedAddress = '0005'
ChangeRequest = '0003'
BindRequestMsg = '0001'
dictValToMsgType = {"0101": "BindResponseMsg"}

# Unexpected error during NAT determination -- try again.
ChangedAddressError = 12

# Requests.
changeRequest = ''.join(
    [ChangeRequest, '0004', "00000006"]
)
changePortRequest = ''.join(
    [ChangeRequest, '0004', "00000002"]
)

def get_stun_servers(af, proto):
    if proto == UDP:
        servers = STUND_SERVERS
    else:
        servers = STUNT_SERVERS

    return servers[af]

# Make a random transaction ID -- using in bind requests.
# Kind of like a poor-mans sequence number for UDP packets.
# Not that useful when TCP is used with the servers.
def gen_tran_id():
    return rand_b(16)

# Filter all other messages that don't match this.
def tran_info_patterns(src_tup=None):
    tranid = gen_tran_id()
    b_msg_p = b".{4}" + re.escape(tranid)
    b_addr_p = b"%s:%d" % (
        re.escape(
            to_b(src_tup[0])
        ),
        src_tup[1]
    )

    return [b_msg_p, b_addr_p, tranid]

# Extract either IPv4 or IPv6 addr from STUN attribute.
def extract_addr(buf, af=socket.AF_INET, base=20):
    # Config for ipv4 and ipv6.
    if af == socket.AF_INET:
        seg_no = 4
        seg_size = 1
        delim = "."
        form = lambda x: str(int(to_h(x), 16))
    else:
        seg_no = 8
        seg_size = 2
        delim = ":"
        form = lambda x: str(to_h(x))

    # Port part.
    p = 6
    port = int(to_h(buf[base + p:base + p + 2]), 16)
    p += 2

    # Binary encodes pieces of address.
    segments = []
    for i in range(0, seg_no):
        part = form(buf[base + p:base + p + seg_size])
        segments.append(part)
        p += seg_size

    # Return human readable version.
    ip = delim.join(segments)
    return [ip, port]

# Check whether a STUN reply is correct.
def stun_check_reply(dest_addr, reply, lax=0):
    try:
        if reply is None:
            return "> STUN reply is none."

        # They replied, go to next test.
        if not reply['resp']:
            return "> STUN get_nat_type first nat test failure"

        # Sanity checking.
        check_fields = [
            'rip', 'rport',
        ]
        for field in check_fields:
            if reply[field] is None:
                return "STUN field %s is None, try again" % (field)

        # External port must be valid port.
        if not valid_port(reply['rport']):
            return "STUNs external port was invalid"

        # Additional fields to check.
        if not lax:
            check_fields = [
                'cip', 'cport',
            ]
            for field in check_fields:
                if reply[field] is None:
                    return "STUN field %s is None, try again" % (field)

            try:
                # If the servers changed address matches its source
                # then it only has one IP and can't be used for the full tests
                if reply["sip"] is not None:
                    if ip_f(reply['cip']) == ip_f(reply["sip"]):
                        return "STUN server only had one address"

                # STUN servers changed IP was the same as the destination address.
                if ip_f(reply['cip']) == ip_f(dest_addr.target()):
                    return "STUN server only had one address"
            except Exception:
                return "STUN ip invalid when given to ip_f"

            # Some STUN servers return local addresses for the changed address.
            # They could be either malicious or misconfigured.
            if not IS_DEBUG:
                if ipaddress.ip_address(reply['cip']).is_private:
                    return "STUNs changed address is private"

            # Filter out this junk IP.
            if reply["cip"] == "8.8.8.8":
                return "STUN change ip was 8.8.8.8"

            # Changed port must be valid port.
            if not valid_port(reply['cport']):
                return "STUNs changed port was invalid"

            # Must be unique to differentiate port tests.
            if reply['cport'] == dest_addr.port:
                return "STUNs changed port same as dest"

        return 0
    except Exception as e:
        return "STUNs check reply unknown error e = %s" % (str(e))

# Send a valid stun request to a server
# and process the reply.
async def do_stun_request(pipe, dest_addr, tran_info, extra_data="", changed_addr=None, conf=STUN_CONF):
    assert(dest_addr is not None)
    ret = {
        # Got response or not.
        'resp': False,

        # Our external IP.
        'rip': None,

        # Our externally mapped port.
        'rport': None,

        # Our local port.
        'lport': pipe.sock.getsockname()[1],

        # Our local IP.
        'lip': pipe.sock.getsockname()[0],

        # Servers IP address.
        'sip': None,

        # Servers reply port.
        'sport': None,

        # IP server will send change requests from.
        'cip': None,

        # Port server will send change requests from.
        'cport': None
    }
    #print(ret)

    # Sanity checking.
    if pipe is None:
        log("> STUN skipping get nat port mapping - s = None")
        return ret

    # Init values.
    str_len = to_hs( struct.pack("!h", int(len(extra_data) / 2)) )
    str_data = ''.join([BindRequestMsg, str_len, to_h(tran_info[2]), extra_data])
    data = binascii.a2b_hex(str_data)
    recvCorr = False
    recieved = False

    # Convenience function to call on failure or complete.
    def handle_cleanup(ret, msg="ok"):
        log("> STUN do request status = %s" % (msg))

        if msg != "ok":
            ret['resp'] = False

        return ret

    # Keep trying to get through.
    log("> STUN do request dest = {}:{}".format(*dest_addr.tup))
    for i in range(0, conf["packet_retry"]):
        try:
            # Send request.
            # Multiplexed for UDP stream.
            # For TCP the dest addr arg is ignored.
            if not await pipe.stream.send(data, dest_addr.tup):
                log("STUN req send all unknown error.")
                continue

            # Receive response -- but only
            # expect a certain client addr.
            buf = await pipe.recv(
                tran_info[:2],
                timeout=conf["recv_timeout"]
            )

            # Error or timeout.
            if buf is None:
                raise asyncio.TimeoutError("STUN recv n timeout.")

            # Check buffer is min length to avoid overflows.
            if len(buf) < 20:
                log("Invalid buf len in main STUN res.")
                continue

            # Only support certain message type.
            msgtype = to_h(buf[0:2])
            if msgtype not in dictValToMsgType:
                log("> STUN unknown msg type %s" % (to_s(msgtype)))
                continue

            # Process response
            # Only interested in bind responses tho.
            bind_resp_msg = dictValToMsgType[msgtype] == "BindResponseMsg"
            tranid_match = tran_info[2] == buf[4:20]
            if bind_resp_msg and tranid_match:
                # Extract length of message attributes.
                len_message = int(to_h(buf[2:4]), 16)
                len_remain = len_message

                # Avoid overflowing buffer.
                if len(buf) - 20 < len_message:
                    log("> Invalid message length recv for stun reply.")
                    continue

                # Start processing message attributes.
                log("> stun parsing bind = %d" % len_remain)
                base = 20
                recvCorr = True
                ret['resp'] = True
                while len_remain > 0:
                    # Avoid overflow for attribute parsing.
                    if base + 4 >= len(buf):
                        log("> new attr field overflow")
                        break

                    # Extract attributes from message buffer.
                    attr_type = to_h(buf[base:(base + 2)])
                    attr_len = int(to_h(buf[(base + 2):(base + 4)]), 16)

                    # Avoid attribute overflows.
                    if attr_len <= 0:
                        log("> STUN attr len")
                        break
                    if attr_len + base + 4 > len(buf):
                        log("> attr len overflow")
                        break

                    # Log attribute type.
                    log("> STUN found attribute type = %s" % (to_s(attr_type)))

                    # Your remote IP and reply port. The important part.
                    if attr_type == MappedAddress:
                        ip, port = extract_addr(buf, dest_addr.chosen, base)
                        ret['rip'] = ip
                        ret['rport'] = port
                        log(f"set mapped address {ip}:{port}")

                    # Original address of the server. Not really that important.
                    if attr_type == SourceAddress:
                        ip, port = extract_addr(buf, dest_addr.chosen, base)
                        ret['sip'] = ip
                        ret['sport'] = port
                        log(f"set SourceAddress {ip}:{port}")

                    # Address that the STUN server would send change reqs from.
                    if attr_type == ChangedAddress:
                        ip, port = extract_addr(buf, dest_addr.chosen, base)
                        ret['cip'] = ip
                        ret['cport'] = port
                        log(f"set ChangedAddress {ip}:{port}")

                    base = base + 4 + attr_len
                    len_remain -= (4 + attr_len)

                break
            else:
                log("> not bind respond msg and tran id match %s %s %s" % (
                    to_s(dictValToMsgType[msgtype]),
                    to_h(tran_info[2]),
                    to_h(buf[4:20]).upper()
                )
                )
        except asyncio.TimeoutError as e:
            log("> STUN get nat port map.. sendto e = %s, timeout = %s" % (str(e), str(conf["recv_timeout"])))

        # Allow other coroutines to do work.
        await asyncio.sleep(0.5)

    return handle_cleanup(ret)    

# Build new pipe or return existing ones
# based on proto. Handle initialization.
async def init_pipe(dest_addr, interface, af, proto, source_port, local_addr=None, conf=STUN_CONF):
    # Make a new socket if this is the first time running.
    assert(dest_addr is not None)

    """
    # Will throw address reuse errors until the port
    # becomes available again. Keep trying to bind until
    # it succeeds -- there's no other way for TCP.
    # This makes determining NAT type via TCP much slower than
    # for UDP. Especially considering multiple tests are done.
    """
    if local_addr is None:
        route = interface.route(af)
        """
        ipr = get_pub_ipr_from_list(route.nic_ips)
        if ipr is not None:
            ips = ipr_norm(ipr)
        else:
            ips = route.nic()
        """

        local_addr = await route.bind(source_port)

    pipe = await pipe_open(
        route=local_addr,
        proto=proto,
        dest=dest_addr,
        conf=conf
    )

    return pipe   

async def stun_check_addr_info(stun_host, stun_port, af, proto, interface, local_addr=None):
    # Load details from local_addr.
    if local_addr is not None:
        route = local_addr
    else:
        route = interface.route(af)

    # Resolve host to ips.
    try:
        dest_addr = await Address(
            stun_host,
            stun_port,
            route,
            proto,
        ).res()
    except Exception as e:
        # Unable to find A or AAA record for address family.
        log("> STUN get_nat_type can't load A/AAA %s" % (str(e)))
        return None

    # No IP address returned for server.
    if dest_addr.target() is None:
        log("> STUN get_nat_type domain a records empty")
        return None

    # If the stun server resolves to a local address then skip it.
    if not IS_DEBUG:
        if ipaddress.ip_address(dest_addr.target()).is_private:
            log("> STUN get_nat_type skipping private dns ip")
            return None

    return dest_addr

# Code for doing a single NAT test with logging.
async def stun_sub_test(msg, dest, interface, af, proto, source_port, changed, extra="", pipe=None, tran_info=None, local_addr=None, conf=STUN_CONF):
    log("> STUN %s" % (msg))

    # Set transaction ID if no match function provided.
    if tran_info is None:
        tran_info = tran_info_patterns(changed.tup)

    # New con for every req if it's TCP.
    if proto == TCP:
        if pipe is not None:
            await pipe.close()
            pipe = None

    # Build sock if none provided.
    if pipe is None:
        # Get new sock with a timeout.
        # The timeout is really only useful for TCP connect.
        pipe = await init_pipe(
            dest,
            interface,
            af,
            proto,
            source_port, 
            local_addr=local_addr,
            conf=conf
        )

        # Check for succcess or not.
        if pipe is None:
            return None, None

    pipe.subscribe(tran_info[:2])
    return await async_wrap_errors(
        do_stun_request(
            pipe,
            dest,
            tran_info,
            extra,
            changed_addr=changed,
            conf=conf
        )
    ), pipe

# Basic interface to the most useful functions.
class STUNClient():
    def __init__(self, interface, af=IP4, sock_timeout=2, consensus=[1, 1], proto=UDP):
        self.interface = interface
        self.af = af
        self.proto = proto
        self.sock_timeout = sock_timeout

        # Threshold / Number = agreement.
        self.t, self.n = consensus

    def rand_server(self):
        servers = get_stun_servers(self.af, self.proto)
        return random.choice(servers)

    """
    Internal function - don't use directly.
    I liek the cat.

    This function has two main modes:
    - regular
    - fast fail

    Regular is designed to use UDP and it will have access to
    far more servers. But the disadvantage is when using
    interfaces that you don't know have a route to the Internet
    you won't know straight away if it works.

    Fast fail uses TCP and will fail immediately if it can't
    do N connections in a row to different servers. This is
    very useful for rapidly determining whether an interface
    is routable via a certain 'address type'.
    """
    async def _get_field(self, name, af, proto, interface, source_port=0, group="map", do_close=0, fast_fail=0, servers=None, local_addr=None, conf=STUN_CONF):
        # Record start time and define fail result funv.
        random.seed()
        start_time = time.time()
        f_fail = lambda: [None, None, time.time() - start_time]
        use_proto = TCP if fast_fail else proto
        lax = 0 if group == "change" else 1

        # Limit at 20 to avoid massive delays.
        # If 20 consecutive servers fail
        # something else is wrong.
        servers = servers or get_stun_servers(af, proto)
        stun_addr = None
        for server in servers:
            for i in range(conf["retry_no"]):
                # Get a valid STUN Address.
                for j in range(conf["addr_retry"]):
                    # Resolve address with a timeout.
                    try:
                        stun_addr = await asyncio.wait_for(
                            stun_check_addr_info(
                                server["primary"]["ip"],
                                server["primary"]["port"],
                                af,
                                proto,
                                interface,
                                local_addr
                            ),
                            conf["dns_timeout"]
                        )
                    except asyncio.TimeoutError:
                        log("> stun addr_task timeout in _getfield")
                        stun_addr = None
                        continue

                    # Check address was resolved.
                    if stun_addr is None:
                        log("> get field error stun addr is None")
                        continue

                    break

                # Monitor basic loop exit.
                if stun_addr is None:
                    continue

                # Do stun test.
                msg = "doing stun sub test for %s" % (name)
                ret = await stun_sub_test(msg, stun_addr, interface, af, proto, source_port, stun_addr, "", local_addr=local_addr, conf=conf)
                nat_info, pipe = ret

                # Check response.
                error = stun_check_reply(stun_addr, nat_info, lax)
                stop_time = time.time() - start_time
                if error:
                    log("> get field error = %s" % (str(error)))
                    if pipe is not None:
                        log("> closing stream get field")
                        await pipe.close()
                        pipe = None

                    continue

                # Return IP section of STUN reply.
                if do_close:
                    if pipe is not None:
                        await pipe.close()
                        pipe = None

                # Return results.
                stop_time = time.time() - start_time
                if name == "nat_info":
                    # Valid STUN reply has main fields set.
                    stun_fields = ["rip", "rport", 'lport', 'lip']
                    do_continue = False
                    for stun_field in stun_fields:
                        if nat_info[stun_field] is None:
                            do_continue = True
                            break

                    # Field wasn't properly set.
                    if do_continue:
                        continue
                    else:
                        return nat_info, pipe, stop_time
                else:
                    # Field wasn't set in STUN reply.
                    if nat_info[name] is None:
                        continue
                    else:
                        return nat_info[name], pipe, stop_time

        # Retry failed.
        return f_fail()

    async def get_wan_ip(self, af=None, interface=None, fast_fail=0, servers=None, local_addr=None, conf=STUN_CONF):
        # Defaults.
        af = af or self.af
        interface = interface or self.interface
        group = "map"

        # sock_timeout + interface added to be compatible
        # with other get wan ip functions -- ignored.
        log("> STUN trying to get WAN IP; af = %d, fast fail = %d" % (af, fast_fail))
        threshold_n = conf["consensus"][0] or self.n
        threshold_t = conf["consensus"][1] or self.t

        # Make list of tests for getting remote IP.
        tasks = []
        for i in range(threshold_n):
            tasks.append(
                self._get_field("rip", af, self.proto, interface, 0, group=group, do_close=1, fast_fail=fast_fail, servers=servers, local_addr=local_addr, conf=conf)
            )

        # Return threshold results.
        f_filter = lambda r: [x for x, _, _ in r if isinstance(x, str)]
        wan_ip = await threshold_gather(tasks, f_filter, threshold_t)
        if wan_ip is not None:
            return ip_norm(wan_ip)

    """
    Coming to consensus from multiple STUN results is not supported for
    get_mapping as it is expected for there to be differences in
    successive mappings as part of how NATs work.
    """
    async def get_mapping(self, proto, af=None, source_port=0, group="map", do_close=0, fast_fail=0, servers=None, conf=STUN_CONF):
        # Defaults.
        af = af or self.af
        log("> stun mapping for proto = %d; af = %d" % (proto, af))

        # Get port mapping from first working server.
        interface = self.interface
        nat_info, s, run_time = await self._get_field("nat_info", af, proto, interface, source_port, group, do_close, fast_fail, servers=servers, conf=conf)
        if nat_info is None:
            return [None, None, None, None, None, None]

        # Return mapping section of STUN reply.
        local = nat_info["lport"] or source_port
        mapped = nat_info["rport"]
        rip = nat_info["rip"]
        if mapped:
            mapped = int(mapped)

        return [interface, s, local, mapped, rip, run_time]
    
#######################################################################

if __name__ == "__main__": # pragma: no cover
    """

    Some custom STUN servers don't include the rip attribute
    for change requests -- in this case we issue get_wan_ips
    to both test addresses and use the result for the tests.
    for ip_check in [[1, stun_addr], [3, nat_tests[1][2]]]:
        n, addr = ip_check
        if test[n]['rip'] is None:
            ret, test_pipe = await stun_sub_test(
                "custom get ip for non conforming server",
                addr,
                interface,
                af,
                proto,
                0,
                addr,
                ""
            )

            print(ret)

            await test_pipe.close()
            test[n]['rip'] = ret['rip']

    # Filters out invalid servers based on response.
    async def get_valid_stun_servers(interface, af, proto, group="map", sock_timeout=2, check_change=0):
        return
        valid_servers = []
        extended_servers = []
        stun_servers = get_stun_servers(af, proto, group)
        for stun_server in stun_servers:
            dest_addr = await stun_check_addr_info(
                stun_server[0],
                stun_server[1],
                af,
                proto,
                stun_check_addr_info
            )
            if dest_addr is None:
                log("> STUN valid servers ... dest addr is none.")
                continue

            s, _ = await init_socket(dest_addr, interface, af, proto, 0)
            if s is None:
                log("> STUN valid servers ... first s is none.")
                continue

            # Get initial port mapping.
            # A response is expected.
            tran_info = tran_info_closure(dest_addr)
            s.subscribe(tran_info[1])
            nat_info = ret = await do_stun_request(
                s,
                dest_addr,
                tran_info
            )

            # Set source port.
            lax = not check_change
            error = stun_check_reply(dest_addr, ret, lax)
            if error:
                log("> STUN valid servers ... first reply error = %s." % (error))
                continue

            if check_change:
                changed_addr = await stun_check_addr_info(
                    ret["ChangcipedIP"],
                    dest_addr.port,
                    af,
                    proto,
                    interface
                )
                if changed_addr is None:
                    log("> STUN valid servers ... changed addr is none.")
                    continue

                s, _ = await init_socket(changed_addr, interface, af, proto, 0)
                if s is None:
                    log("> STUN valid servers ... second s is none.")
                    continue

                tran_info = tran_info_closure(changed_addr)
                s.subscribe(tran_info[1])
                ret = await do_stun_request(
                    s,
                    changed_addr,
                    tran_info
                )

                # Check reply.
                error = stun_check_reply(changed_addr, ret, lax)
                if error:
                    log("> STUN valid servers ... error 2 = %s." % (error))
                    continue

            extended_servers.append(
                [stun_server[0], nat_info["cport"]]
            )
            extended_servers.append(
                [ret["cip"], ret["cport"]]
            )

            valid_servers.append([stun_server[0], stun_server[1]])

        return valid_servers, extended_servers
    """

    async def test_stun_client():

        from .interface import Interface, load_interfaces, init_p2pd

        return

        # Maybe has a turn server.
        # https://wiki.innovaphone.com/index.php?#title=Howto:Innovaphones_public_services#TURN
        server = ['stun.innovaphone.com', 3478]



        i = await Interface().start_local()
        s = STUNClient(interface=i, proto=UDP, af=IP6)
        x = await s.get_wan_ip()
        print(x)
        return

        out = await s.get_mapping(proto=TCP)
        print(out)
        return

        i = await Interface().start()
        print(i)

        return



        """
        if there's no default route for an af then the route
        func should skip it.
        """

        """
        r = interface.route()
        await r.bind()
        print(r.bind.ip)
        print(r.bind.port)
        return
        """

        stun_client = STUNClient(
            interface,
            af,
            consensus=[1, 1]
        )


        # Which bind IP for ip4?
        # How to do the open internet check?


        route = await interface.rp[af].routes[0].bind()
        ret = await stun_client.get_wan_ip(local_addr=route)
        #ret = await stun_client.get_nat_type()
        print(ret)
        return
        nat = await stun_client.get_nat_info()
        print(nat)
        #print(ret)
        return

        ret = await stun_client.get_mapping(proto)
        print(ret)
        #return

    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    loop.run_until_complete(test_stun_client())
    loop.close()




import functools
import asyncio
import re
import time
import binascii
import ipaddress
import sys, os
import platform
import logging
import traceback
import random
import inspect
#import uvloop
import itertools
import ctypes
import urllib.parse
import platform
import selectors
import copy
import hashlib

# Yoloswaggins.
if not hasattr(asyncio, 'create_task'):
    asyncio.create_task = asyncio.ensure_future

vmaj, vmin, _ = platform.python_version_tuple()
vmaj = int(vmaj); vmin = int(vmin)
if vmaj < 3:
    raise Exception("Python 2 not supported.")
if vmin < 8:
    if sys.platform == 'win32':
        #raise Exception("Windows needs Python 3.8 or higher.")
        pass
if vmin < 5:
    raise Exception("Non-Windows OS needs Python 3.5 or higher")

def proactorfy(self=None):
    """
    The func is called once to set the main event loop
    to use. At this point there is an existing event loop
    not doing anything. This event loop should be used first
    rather than creating a new one. With aiounittest this
    prevents having unclosed event loops at the end of
    running test suites.
    """
    try:
        loop = asyncio.get_event_loop()
    except:
        loop = asyncio.ProactorEventLoop()

    asyncio.set_event_loop(loop)
    if self is not None:
        self.my_loop = loop
        return self.my_loop

    return loop

if sys.platform == 'win32':
    # Won't work on older Python < 3.6.
    try:
        proactorfy()

        # Only available on later Python versions.
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
    except Exception:
        pass

if "P2PD_DEBUG" in os.environ: 
    IS_DEBUG = 1
    logging.basicConfig(
        filename='program.log',
        level=logging.DEBUG,
        format='[%(asctime)s.%(msecs)03d] @ [%(filename)s:%(lineno)d] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    def log(m):
        if "P2PD_DEBUG" not in os.environ:
            return

        logging.info(m)
else:
    IS_DEBUG = 0
    log = lambda m: None

STATUS_RETRY = 1
STATUS_SUCCESS = 2
STATUS_FAILURE = 3
MAX_PORT = 65535

re.unescape = lambda x: re.sub(r'\\(.)', r'\1', x)
to_b = lambda x: x if type(x) == bytes else x.encode("ascii", errors='ignore')
to_s = lambda x: x if type(x) == str else x.decode("utf-8", errors='ignore')
to_hs = lambda x: to_s(binascii.hexlify(to_b(x)))
to_h = lambda x: to_hs(x) if len(x) else "00"
to_i = lambda x: x if isinstance(x, int) else int(x, 16)
to_n = lambda x: x if isinstance(x, int) else int(to_s(x), 10)
i_to_b = lambda x, o='little': x.to_bytes((x.bit_length() + 7) // 8, o)
b_to_i = lambda x, o='little': int.from_bytes(x, o)
valid_port = lambda p: p >= 1 and p <= MAX_PORT
port_wrap = lambda p: (p % MAX_PORT) or 1
to_unique = lambda x: [i for n, i in enumerate(x) if i not in x[:n]]
ip_f = ipaddress.ip_address
class_name = lambda x: type(x).__name__ if inspect.isclass(x) else None
timestamp = lambda p=0: time.time() if p else int(time.time())
strip_none = lambda x: [i for i in x if i]
neg_flip = lambda r, x, y: -r if x > y else r
n_dist = lambda x, y: neg_flip(max(x, y) - min(x, y), x, y)
dict_plus = lambda d, k: d[k] if k in d else 0
rand_rang = random.randrange
list_join = lambda l: list(itertools.chain.from_iterable(l))
from_range = lambda r: rand_rang(r[0], r[1] + 1)
in_range = lambda x, r: x >= r[0] and x <= r[1]
b_and = lambda abytes, bbytes: bytes(map(lambda a,b: a & b, abytes, bbytes))
b_or = lambda abytes, bbytes: bytes(map(lambda a,b: a | b, abytes, bbytes))
len_range = lambda r: r[1] - r[0]
get_bits = lambda n, l, p=0: ( ((1 << l) - 1)  &  (n >> p ) )
is_no = lambda x: to_s(x).isnumeric()
is_b = lambda x: isinstance(x, bytes)
rm_whitespace = lambda x: re.sub(r"\s+", "", x, flags=re.UNICODE)
urlencode = lambda x: to_b(urllib.parse.quote(x))
urldecode = lambda x: to_b(urllib.parse.unquote(x))
shuffle = lambda x: random.shuffle(x) or x
d_keys = lambda x: list(x.keys())
d_vals = lambda x: list(x.values())
sha256 = lambda x: to_s(hashlib.sha256(to_b(x)).hexdigest())
hash160 = lambda x: hashlib.new('ripemd160', to_b(x)).digest()
sha3_256 = lambda x: to_s(hashlib.sha3_256(to_b(x)).hexdigest())
b_sha3_256 = lambda x: hashlib.sha3_256(to_b(x)).digest()

def list_exclude_dict(key_name, exclusion, entry_list):
    sub_list = []
    for entry in entry_list:
        if key_name not in entry:
            sub_list.append(entry)
            continue

        if entry[key_name] != exclusion:
            sub_list.append(entry)

    return sub_list

def list_get_dict(key_name, criteria, entry_list):
    for entry in entry_list:
        if key_name not in entry:
            continue

        if entry[key_name] == criteria:
            return entry

def file_get_contents(path):
    with open(path, mode='rb') as f:
        return f.read()

def to_type(x, out_type):
    # String.
    if isinstance(out_type, str):
        return to_s(x)

    # Bytes.
    if isinstance(out_type, bytes):
        return to_b(x)

# Take a dict template called Y and a child dict called X.
# Yield a new dict with Y's vals overwritten by X's.
def dict_child(x, y):
    if len(x) > len(y):
        log("dict child x > y, are args in correct order?")

    out = copy.deepcopy(y)
    for key in x:
        out[key] = x[key]
        
    return out

def dict_merge(x, y):
    x.update(y)
    return x

def rand_plain(n):
    charset =  b"012345678abcdefghijklmnopqrs"
    charset += b"tuvxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    buf = b""
    for meow in range(0, n):
        ch = random.choice(charset)
        buf += bytes([ch])
        
    return buf

def ensure_resolved(targets):
    if not isinstance(targets, list):
        targets = [targets]

    for i, target in enumerate(targets):
        if not target.resolved:
            e = "Target offset = {}, id = {}, type = {} not resolved".format(
                i,
                id(target),
                type(target)
            )

            raise Exception(e)

def sorted_search(n_list, i, start_at=None):
    # Make sure list entries exist.
    list_len = len(n_list)
    if not list_len:
        return None

    # [2, 2, 2, 2] even -> index 2 -> right of a pair.
    # [1, 2, 3] odd -> index 1 -> middle, odd one out.
    x = 0
    if start_at is None:
        start_index = int(list_len / 2)
    else:
        start_index = start_at % list_len

    # Find nearest number.
    while x < list_len:
        # Avoid underflow.
        if start_index <= 0:
            return 0

        # X >= i: (match or decrease) by half
        # else < i: (increase) by half
        if n_list[start_index] >= i:
            if n_list[start_index - 1] < i:
                return start_index
            else:
                start_index -= int(start_index / 2) or 1
        else:
            start_index += int(start_index / 2) or 1

        # Reached the end.
        if start_index >= list_len - 1:
            return list_len - 1

        # If items aren't sorted an infinite loop may be possible.
        x += 1

    raise Exception("Sorted search: List may not be sorted.")

async def threshold_gather(tasks, f_filter, t):
    # Do tests concurrently.
    results = await asyncio.gather(*tasks)
    results = strip_none(results)
    results = sorted(f_filter(results))
    if not len(results):
        return None

    # Return first result that satisfies threshold.
    count = 0
    cur = results[0]
    for i in range(0, len(results)):
        if results[i] == cur:
            count += 1
        else:
            count = 1
            cur = results[i]

        if count >= t:
            return cur

    # Threshold requirement not met.
    return None

"""
https://stackoverflow.com/questions/843828/how-can-i-check-hamming-weight-without-converting-to-binary
"""
def hamming_weight(n: int) -> int:
    c = 0
    while n:
        c += 1
        n &= n - 1

    return c

def bits_to_bytes(s):
    return int(s, 2).to_bytes((len(s) + 7) // 8, byteorder='big')

def range_intersects(a, b):
    # If ranges are the same they intersect.
    if a == b:
        return True

    # Make a list of the ranges and make the pairs ascend.
    # Then make a list of all the elements together.
    range_list = sum(sorted([a, b]), [])
    return range_list[2] <= range_list[1]

def intersect_range(a, b):
    # Add all the elements together.
    els = sum([a, b], [])

    # Sort from small to large.
    asc = sorted(els)

    # Middle two els == intersect range.
    return [asc[1], asc[2]]

def rand_b(n):
    buf = b""
    for i in range(0, n):
        buf += bytes([rand_rang(256)])

    return buf

def rand_b_readable(n):
    chars = b"abcdefghijklmnopqrstuvwxyz"
    chars += b"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    chars += b"0123456789 "

    buf = b""
    for _ in range(n):
        buf += bytes([random.choice(chars)])

    return buf

def field_wrap(n, field):
    start_range, stop_range = field
    stop_range += 1
    y = x = n % stop_range
    while 1:
        if x < start_range:
            x += start_range
            y = x % stop_range
            if x != y:
                x = y
        #
        if x == y:
            break
    #
    return x

def field_dist(x, y, field):
    # Get distance between numbers.
    max_no = max(x, y)
    min_no = min(x, y)
    ret = dist = max_no - min_no

    # They are the same value.
    if not dist:
        return 0

    # min is closer to the start of the field.
    rem = field - max_no
    field_dist = min_no + rem
    if field_dist < dist:
        ret = field_dist

    # Calculate sign of result.
    if x == min_no:
        return -ret
    else:
        return ret

def what_exception():
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    print(exc_type, fname, exc_tb.tb_lineno)
    print(traceback.format_exc())

def log_exception():
    exc_type, exc_obj, exc_tb = sys.exc_info()
    fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
    exc_out = traceback.format_exc()
    log("> {}, line {} = {}".format(
        fname,
        exc_tb.tb_lineno,
        exc_out
    ))

async def async_wrap_errors(coro, timeout=None):
    try:
        # Don't bound wait time.
        if timeout is None:
            return (await coro)

        # Bound wait time.
        if isinstance(timeout, int):
            return (await asyncio.wait_for(coro, timeout))
    except Exception as e:
        # Log all errors.
        log("async wrap errors called")
        log_exception()

def sync_wrap_errors(f, args=[]):
    try:
        if len(args):
            return f(*args)
        else:
            return f()
    except Exception:
        # Log all errors.
        try:
            log_exception()
        except Exception:
            pass

async def async_retry(gen, count, timeout=4):
    # Timeout counter and retry counter.
    i = r = 0
    while 1:
        # Wait for success or timeout.
        try:
            # Run main coroutine
            # get retry status and retry coro.
            init_coro = gen()
            status_future, retry_coro, new_future = await asyncio.wait_for(
                async_wrap_errors(init_coro),
                timeout
            )

            # Do first retry coroutine.
            if i == 0:
                await retry_coro()

            # Get return status after first retry.
            status = await asyncio.wait_for(status_future, timeout)

            # Overwrite old future so it can be set again.
            status_future = new_future()

            # Schedule another attempt.
            if status == STATUS_RETRY:
                i -= 1
                r += 1
                await retry_coro()

            # Success or failure.
            if status != STATUS_RETRY:
                return None

        # Timeout occured -- try again.
        except asyncio.TimeoutError:
            pass

        # Limit retries to no.
        finally:
            i += 1
            if count in (i, r):
                break

    raise asyncio.TimeoutError("Async retry timeout")

def rm_done_tasks(hey):
    pending_tasks = []
    for task in hey:
        if not task.done():
            pending_tasks.append(task)

    return pending_tasks

# Process return value.
def handler_done_builder(pipe, handler, task=None):
    def closure(result):
        # Remove task from pipe.
        if task in pipe.tasks:
            pipe.handler_tasks.remove(task)

        # Got an int result -- check if its error code.
        if isinstance(result, int):
            # Error code.
            if result:
                # Log the error.
                out = f"> {handler} = error {result}."
                log(out)

        # If it returns a task then save it.
        if isinstance(result, asyncio.Task):
            pipe.tasks.append(result)

    return closure

def run_handler(pipe, handler, client_tup, data=None):
    # It's async.
    if inspect.iscoroutinefunction(handler):
        # Lets you process messages from an async func.
        task = asyncio.create_task(
            async_wrap_errors(
                handler(data, client_tup, pipe)
            )
        )

        # Process result if anyone.
        task.add_done_callback(
            handler_done_builder(pipe, handler, task)
        )

        # Needed or they might be garbage collected.
        pipe.handler_tasks.append(task)
    else: 
        # It's a callback.
        result = sync_wrap_errors(
            handler, [data, client_tup, pipe]
        )

        # Process result if any.
        handler_done_builder(pipe, handler)(result)

# Used for event-based programming.
# Can execute code on new cons, dropped cons, and new msgs.
def run_handlers(pipe, handlers, client_tup, data=None):
    # Run any registered call backs on msg.
    pipe.handler_tasks = rm_done_tasks(pipe.handler_tasks)
    for handler in handlers:
        # Run the handler as a callback or coroutine.
        run_handler(pipe, handler, client_tup, data)

def run_in_executor(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        if not inspect.iscoroutinefunction(f):
            loop = asyncio.get_event_loop()
            return loop.run_in_executor(None, lambda: f(*args, **kwargs))
        else:
            def helper():
                loop = asyncio.new_event_loop()
                try:
                    coro = f(*args, **kwargs)
                    asyncio.set_event_loop(loop)
                    return loop.run_until_complete(coro)
                finally:
                    loop.close()

            loop = asyncio.get_event_loop()
            return loop.run_in_executor(None, helper)

    return inner

def cancel_tasks(tasks):
    for task in tasks:
        task.cancel()

# Wait for tasks to finish up until a timeout.
# Otherwise cancel them and wait for errors.
async def gather_or_cancel(tasks, timeout):
    group_task = asyncio.gather(*tasks)
    try:
        await asyncio.wait_for(
            group_task,
            timeout
        )
    except asyncio.TimeoutError:
        for task in tasks:
            task.cancel()

        try:
            await group_task
        except Exception:
            pass
        await asyncio.sleep(0)
    except asyncio.CancelledError:
        return []
    finally:
        return []

class SelectorEventPolicy(asyncio.DefaultEventLoopPolicy):
    """
    @staticmethod
    def exception_handler(self, context):
        print("exception handler")
        print(context)
    """

    @staticmethod
    def loop_setup(loop):
        loop.set_debug(False)
        #loop.set_exception_handler(SelectorEventPolicy.exception_handler)
        #loop.default_exception_handler = SelectorEventPolicy.exception_handler

    def new_event_loop(self):
        selector = selectors.SelectSelector()
        loop = asyncio.SelectorEventLoop(selector)
        SelectorEventPolicy.loop_setup(loop)
        return loop

def selector_event_loop():
    selector = selectors.SelectSelector()
    return asyncio.SelectorEventLoop(selector)

def get_loop(loop=None):
    if loop is None:
        if sys.platform == "win32":
            loop = asyncio.ProactorEventLoop()
        else:
            loop = asyncio.get_event_loop()
    else:
        loop = loop()

    return loop

# Handle stray exceptions in the event loop.
def handle_exceptions(loop, context):
    pass

# Will be used in sample code to avoid boilerplate.
def async_test(f, args=[], loop=None):
    #uvloop.install()
    loop = get_loop(loop)
    #if IS_DEBUG:
    #    loop.set_debug(True)
    loop.set_debug(True)

    # Can have cleanup errors.
    if len(args):
        loop.run_until_complete(f(*args))
    else:
        loop.run_until_complete(f())
    loop.close()
 

async def return_true(result=None):
    return True
            
# If there's an error that its already in a loop
# Run nest_asyncio.apply()
def async_to_sync(f, params=None, loop=None):
    loop = loop or get_loop()
    if params is not None:
        def closure(args):
            return loop.run_until_complete(f(*args))
            
        return closure
    else:
        def closure():
            return loop.run_until_complete(f())
            
        return closure

def get_running_loop():
    try:
        version = sys.version_info[1]
        if version >= 7:
            return asyncio.get_running_loop()
        else:
            return asyncio.get_event_loop()
    except RuntimeError:
        return None

def sqlite_dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d

if __name__ == "__main__": # pragma: no cover
    x = [1, 1]
    y = [2, 2]
    assert(range_intersects(x, y) == False)

    x = [1, 2]
    y = [2, 2]
    assert(range_intersects(x, y) == True)

    x = [1, 2]
    y = [1, 2]
    assert(range_intersects(x, y) == True)

    
    x = [1, 10]
    y = [5, 20]
    assert(range_intersects(x, y) == True)

    



from p2pd import *

async def example():
    i = await Interface().start()
    addr = await Address("www.example.com", 80, i.route())
    curl = WebCurl(addr)
    url_params = {"q": "lets search!"}
    resp = await curl.vars(url_params).get("/")
    print(resp.out)

if __name__ == '__main__':
    async_test(example)



import sys
import os
import asyncio

"""
This is a hack to avoid double-imports of a module when using
the -m switch to run a module directly. Python modules are lolz.
"""
if not '-m' in sys.argv:
    os.environ["PYTHONIOENCODING"] = "utf-8"
    from .errors import *
    from .utils import log, what_exception, log_exception, async_test, SelectorEventPolicy
    
    from .cmd_tools import *
    from .net import *
    from .address import Address
    from .ip_range import IPRange
    from .upnp import port_forward
    from .route import Route, RoutePool, Routes
    from .base_stream import pipe_open, SUB_ALL
    from .interface import Interface, Interfaces, init_p2pd
    from .clock_skew import SysClock
    from .stun_client import STUNClient
    from .turn_client import TURNClient
    from .tcp_punch import TCPPunch
    from .daemon import Daemon
    from .http_client_lib import http_req, ParseHTTPResponse
    from .http_client_lib import http_req_buf
    from .http_server_lib import rest_service, send_json, send_binary, RESTD, api_route_closure
    from .http_server_lib import ParseHTTPRequest
    from .rest_api import P2PDServer, start_p2pd_server
    from .p2p_addr import *
    from .p2p_pipe import *
    from .p2p_node import P2PNode
    from .p2p_utils import get_pp_executors, start_p2p_node
    from .install import *
    from .toxiclient import ToxiToxic, ToxiTunnel, ToxiClient
    from .toxiserver import ToxiMainServer
    from .irc_dns import *

if __name__ == "__main__":
    multiprocessing.set_start_method("spawn")


import asyncio
import json
import re
import random
from .test_init import *
from .daemon import Daemon
from .http_server_lib import RESTD

class ToxicBase():
    def __init__(self, name=None, direction=None, toxicity=None):
        self.name = name
        self.direction = direction
        self.toxicity = toxicity
    def setup(self, base):
        self.name = base.name
        self.direction = base.direction
        self.toxicity = base.toxicity
        return self
    def should_run(self):
        p = min(int(self.toxicity * 100), 100)
        r = random.randrange(1, 101)
        if r <= p:
            return True
        else:
            return False
    async def placeholder(self, msg, dest_pipe):
        return msg, dest_pipe
        
async def toxic_router(msg, src_pipe, dest_pipe, toxics):
    for toxic in toxics:
        if toxic.should_run():
            ret = await async_wrap_errors(
                toxic.run(msg, dest_pipe)
            )

            if ret is None:
                continue
            else:
                msg, dest_pipe = ret

            # Dest pipe has been closed.
            if dest_pipe is None:
                return
            
            # Message filtered.
            if msg is None:
                return

    await async_wrap_errors(
        dest_pipe.send(msg)
    )

class ToxicLatency(ToxicBase):
    def __init__(self):
        super().__init__(self)
    def set_params(self, latency, jitter):
        self.latency = latency
        self.jitter = jitter
        return self
    async def run(self, msg, dest_pipe):
        # Simulate different message arrival times.
        jitter = 0
        if self.jitter:
            jitter = random.randrange(0, self.jitter)
            if random.choice([0, 1]):
                jitter = -jitter
        # Return control to other coroutines.
        ms = self.latency + jitter
        if ms > 0:
            await asyncio.sleep(ms / 1000)
        return msg, dest_pipe
    
class ToxicBandwidthLimit(ToxicBase):
    def set_params(self, rate):
        self.rate = rate # KB/s

        # Track amount of bw used across coroutines.
        # This prevents exceeding the bw allocation.
        self.used = 0

        # Convert rate to bytes per 100 ms.
        # If rate is 0 then there's no limit.
        self.rate_ms = int((self.rate * 1024) / 10)
        return self
    
    async def get_bandwidth(self, n):
        # Wait for some positive amount of bw to become available.
        # Other coroutines may be using all the bw.
        bw = 0
        while bw <= 0 and self.rate_ms:
            bw = self.rate_ms - self.used

            """
            Allocations are recorded before doing a sleep since this
            allows other coroutines to run which may then get their
            own allocations and end up going over the limit. Doing
            the change immediately after ensures the bandwidth limit
            is respected between coroutines.
            """
            if bw > 0:
                # Use only as much as needed.
                bw = min(n, bw)
                self.used += bw

            """
            Give other coroutines a chance to use their bw.
            Sleeping is always done to start with even if
            there was initial bw available because otherwise
            multiple coroutines could accidentally exceed the limit.
            """
            await asyncio.sleep(0.1)

        # Return the allocated amount.
        return bw

    # Could be improved with memoryview.
    async def run(self, msg, dest_pipe):
        unsent = len(msg)
        while unsent > 0:
            # Block for an allocation of bandwidth.
            # Checks every 100 ms.
            bw = await self.get_bandwidth(unsent)

            # Once an allocation is given use that amount.
            await dest_pipe.send(msg[:bw])
            msg = msg[bw:]

            # Adjust counters after await operations.
            # This is for safety to maintain integrity of self.used.
            unsent -= bw
            self.used -= bw

        # Message sent so return.
        return None, dest_pipe

class ToxicSlowClose(ToxicBase):
    def set_params(self, ms):
        self.ms = ms
        return self

    async def run(self, msg, dest_pipe):
        # Close is already patched so undo.
        # Allows for later slow closes to replace this.
        if hasattr(dest_pipe, "regular_close"):
            dest_pipe.close = dest_pipe.regular_close

        # Save old close function.
        dest_pipe.regular_close = dest_pipe.close

        # New function to add slow close.
        async def slow_close():
            await asyncio.sleep(self.ms / 1000)
            await dest_pipe.regular_close()

        # Replace old close with slow close.
        dest_pipe.close = slow_close

        # Replace this class function instance
        # so it only runs once.
        self.run = self.placeholder

class ToxicTimeout(ToxicBase):
    def set_params(self, ms):
        self.start_time = time.time()
        self.ms = ms
        if self.ms:
            self.ms = ms / 1000.0

        return self

    async def run(self, msg, dest_pipe):
        # Calculate duration of active toxic.
        cur_time = time.time()
        duration = cur_time - self.start_time

        # Close if timeout reached.
        if self.ms:
            remaining = self.ms - min(self.ms, duration)
            if remaining:
                await asyncio.sleep(remaining)

            await dest_pipe.close()
            return None, None

        # Block data transmission.
        return None, dest_pipe
    
class ToxicResetPeer(ToxicBase):
    def set_params(self, ms):
        self.start_time = time.time()
        self.ms = ms
        if self.ms:
            self.ms = ms / 1000.0
            
        return self

    async def run(self, msg, dest_pipe):
        # Discard unacked data on close.
        if not hasattr(dest_pipe, "reset_peer_hook"):
            dest_pipe.reset_peer_hook = 1
            linger = struct.pack('ii', 1, 0)
            dest_pipe.sock.setsockopt(
                socket.SOL_SOCKET,
                socket.SO_LINGER,
                linger
            )

        # Calculate duration of active toxic.
        cur_time = time.time()
        duration = cur_time - self.start_time

        # Close if timeout reached.
        if self.ms:
            if duration >= self.ms:
                await dest_pipe.close()
            else:
                return msg, dest_pipe
        else:
            await dest_pipe.close()

        # Block data transmission.
        return None, None

class ToxicLimitData(ToxicBase):
    def set_params(self, n):
        self.n = n
        self.total = 0
        return self

    async def run(self, msg, dest_pipe):
        self.total += len(msg)
        if self.total >= self.n:
            await dest_pipe.close()
            return None, None
        else:
            return msg, dest_pipe
        
class ToxicSlicer(ToxicBase):
    def set_params(self, avg_size, size_var, delay):
        self.avg_size = avg_size
        self.size_var = size_var
        self.delay = delay
        return self
    
    def chunk(self, start, end):
        """
        Base case:
        If the size is within the random variation,
        or already less than the average size, just
        return it. Otherwise split the message into
        chunks of average size +/- size variation.
        """
        offsets = []
        seg_len = end - start
        if seg_len - self.avg_size <= self.size_var:
            return [start, end]
        
        # Build chunk offset list.
        # End offset overlaps with start offset.
        # This is not a mistake.
        p_start = start
        while 1:
            # Calculate a random variation.
            # May be positive or negative.
            rand_len = rand_rang(0, (self.size_var * 2) + 1)
            change = rand_len - self.size_var

            # Increase start pointer by random variation.
            # Avoid underflows with max.
            p_end = p_start + max((self.avg_size + change), 1)

            # Record the chunk offsets.
            # Avoid overflows with min.
            offsets += [p_start, min(p_end, end)]

            # Increase pointer for next segment.
            p_start = offsets[-1]

            # Should never be greater than end
            # but this will be used as the cutoff.
            if offsets[-1] >= end:
                break

        return offsets

    async def run(self, msg, dest_pipe):
        offsets = self.chunk(0, len(msg))
        if self.delay:
            await asyncio.sleep(self.delay / 1000)

        for i in range(1, len(offsets), 2):
            chunk = msg[offsets[i - 1]:offsets[i]]
            if self.delay:
                await asyncio.sleep(self.delay / 1000)

            await dest_pipe.send(chunk)

        # Messages are processed so return None.
        return None, dest_pipe

# You need a way to force new messages to wait for past.
class ToxiTunnelServer(Daemon):
    def __init__(self, name):
        super().__init__()
        self.name = name
        self.upstream_pipe = None
        self.upstream_toxics = {}
        self.downstream_toxics = {}
        self.clients = []

    def get_toxic(self, name):
        if name in self.upstream_toxics:
            return self.upstream_toxics
        
        if name in self.downstream_toxics:
            return self.downstream_toxics
        
        return None

    async def close(self):
        # Ensure all clients are closed.
        for client in self.clients:
            await client.close()

        # Close connection to upstream if it's set.
        if self.upstream_pipe is not None:
            await self.upstream_pipe.close()

        # Close the listen server itself.
        await super().close()

    # Record a list of tunnel clients to relay messages from upstream to.
    async def up_cb(self, msg, client_tup, con_pipe):
        # Make sure they are removed when the connection closes.
        async def end_cb(msg, client_tup, end_pipe):
            if con_pipe in self.clients:
                self.clients.remove(con_pipe)

        con_pipe.add_end_cb(end_cb)
        self.clients.append(con_pipe)

    # Redirect msgs from upstream to clients with toxics.
    def set_upstream(self, pipe):
        async def msg_cb(msg, client_tup, upstream_pipe):
            pipe.transport.pause_reading()
            tasks = []
            for client in self.clients:
                tasks.append(
                    toxic_router(
                        msg,
                        pipe,
                        client,
                        d_vals(self.downstream_toxics)
                    )
                )

            if len(tasks):
                await asyncio.gather(
                    *tasks,
                    return_exceptions=True
                )

            pipe.transport.resume_reading()

        pipe.add_msg_cb(msg_cb)
        self.upstream_pipe = pipe

    # Redirect msg from clients to upstream with toxics.
    async def msg_cb(self, msg, client_tup, down_pipe):
        down_pipe.transport.pause_reading()
        await toxic_router(
            msg,
            down_pipe,
            self.upstream_pipe,
            d_vals(self.upstream_toxics)
        )
        down_pipe.transport.resume_reading()

class ToxiMainServer(RESTD):
    def __init__(self, interfaces):
        super().__init__()
        self.interfaces = interfaces
        self.tunnel_servs = {}

    @RESTD.POST(["proxies"])
    async def create_tunnel_server(self, v, pipe):
        j = v["body"]

        # This is a request to close a tunnel.
        if j["enabled"] == False:
            # Close server and delete record of tunnel.
            tunnel_name = v["name"]["proxies"]
            tunnel_serv = self.tunnel_servs[tunnel_name]
            await tunnel_serv.close()
            del self.tunnel_servs[tunnel_name]

            # Indicate tunnel is closed.
            return {
                "name": tunnel_name,
                "enabled": False,
                "toxics": []
            }

        # Otherwise it's a new create call.
        tup_pattern = "([\s\S]+):([0-9]+)$"
        bind_ip, bind_port = re.findall(
            tup_pattern,
            j["listen"]
        )[0]

        # Build listen route.
        route = await pipe.route.interface.route().bind(
            ips=bind_ip,
            port=int(bind_port)
        )

        # Start the tunnel server.
        tunnel_serv = ToxiTunnelServer(name=j["name"])
        await tunnel_serv.listen_specific(
            [[route, TCP]]
        )

        # Extract destination IP.
        dest_ip, dest_port = re.findall(
            tup_pattern,
            j["upstream"]
        )[0]

        # rewrite this to use an interface that supports
        # dest AF
        use_if = None
        dest_ipr = IPRange(dest_ip)
        for interface in self.interfaces:
            if dest_ipr.af in interface.supported():
                use_if = interface

        # Could not find suitable interface.
        if use_if is None:
            return {
                "error": "no interfaces for that address family."
            }

        # Resolve upstream address.
        up_route = await use_if.route().bind()
        dest = await Address(dest_ip, dest_port, up_route)

        # Connect to upstream.
        upstream = await pipe_open(
            TCP,
            up_route,
            dest
        )

        # Add upstream to tunnel server.
        tunnel_serv.set_upstream(upstream)

        # If zero was passed convert to port no.
        bind_port = route.bind_port
        self.tunnel_servs[j["name"]] = tunnel_serv

        # Return response
        return {
            "name": j["name"],
            "listen": f"{bind_ip}:{bind_port}",
            "upstream": j["upstream"],
            "enabled": True,
            "toxics": []
        }

    @RESTD.POST(["proxies"], ["toxics"])
    async def add_new_toxic(self, v, pipe):
        j = v["body"]; attrs = j["attributes"]
        proxy_name = v["name"]["proxies"]
        if proxy_name not in self.tunnel_servs:
            return {
                "error": "tunnel name not found"
            }
        
        # Make sure name doesn't exist.
        serv = self.tunnel_servs[proxy_name]
        if serv.get_toxic(j["name"]) is not None:
            return {
                "error": "toxic name already exists."
            }

        # Create toxic.
        toxic = base = ToxicBase(j["name"], j["stream"], j["toxicity"])

        # Clauses for all the toxics.
        if j["type"] == "slow_close":
            toxic = ToxicSlowClose().set_params(
                ms=attrs["delay"]
            ).setup(base)

        if j["type"] == "bandwidth":
            toxic = ToxicBandwidthLimit().set_params(
                rate=attrs["rate"]
            ).setup(base)

        if j["type"] == "timeout":
            toxic = ToxicTimeout().set_params(
                ms=attrs["timeout"]
            ).setup(base)

        if j["type"] == "slicer":
            toxic = ToxicSlicer().set_params(
                avg_size=attrs["average_size"],
                size_var=attrs["size_variation"],
                delay=attrs["delay"]
            ).setup(base)
    
        if j["type"] == "latency":
            toxic = ToxicLatency().set_params(
                latency=attrs["latency"],
                jitter=attrs["jitter"]
            ).setup(base)

        if j["type"] == "reset":
            toxic = ToxicResetPeer().set_params(
                ms=attrs["timeout"]
            ).setup(base)

        if j["type"] == "limit_data":
            toxic = ToxicLimitData().set_params(
                n=attrs["bytes"]
            ).setup(base)

        # Add toxic to tunnel server.
        if j["stream"] == "upstream":
            toxics = serv.upstream_toxics
        else:
            toxics = serv.downstream_toxics
        toxics[j["name"]] = toxic

        # Response.
        return j

    @RESTD.DELETE(["proxies"], ["toxics"])
    async def del_new_toxic(self, v, pipe):
        proxy_name = v["name"]["proxies"]
        toxic_name = v["name"]["toxics"]
        if proxy_name in self.tunnel_servs:
            tunnel = self.tunnel_servs[proxy_name]
            toxics = tunnel.get_toxic(toxic_name)
            if toxics is not None:
                del toxics[toxic_name]

        return b""


    # Ensure all tunnel servers are also closed.
    async def close(self):
        for tunnel_serv in d_vals(self.tunnel_servs):
            await tunnel_serv.close()

        await super().close()

asyncio.set_event_loop_policy(SelectorEventPolicy())

"""
class TestToxiServer(unittest.IsolatedAsyncioTestCase):

    async def test_toxi_server(self):
        #ret = api_route_closure("/proxies/test") ([["proxies"], ["toxics"]])
        #print(ret)
        #return


        got_n, got_p = api_route_closure("/test")([])
        want_n = {}
        want_p = {0: "test"}
        assert(got_n == want_n and got_p == want_p)

        got_n, got_p = api_route_closure("/test")([["test"]])
        want_p = {}
        want_n = {"test": "test"}
        assert(got_n == want_n and got_p == want_p)

        got_n, got_p = api_route_closure("/test/val")([["test"]])
        want_p = {}
        want_n = {"test": "val"}
        assert(got_n == want_n and got_p == want_p)

        got_n, got_p = api_route_closure("/test/val/xx/yy/aa")([["test"], ["yy", "bb"]])
        want_p = {0: 'xx'}
        want_n = {'test': 'val', 'yy': 'aa'}
        assert(got_n == want_n and got_p == want_p)

        got_n, got_p = api_route_closure("/test/val")([["test"]])
        want_n = {"test": "val"}
        want_p = {}
        assert(got_n == want_n and got_p == want_p)

        got_n, got_p = api_route_closure("/a/b")([["a", 0], ["b", 1]])
        want_n = {"a": 0, "b": 1}
        want_p = {}
        assert(got_n == want_n and got_p == want_p)
        #print(got_n, got_p)
        #assert(got == want)

        i = await Interface().start_local(skip_resolve=True)
        route = await i.route().bind(ips="127.0.0.1", port=8475)
        server = ToxiMainServer([i])
        await server.listen_specific(
            [[route, TCP]]
        )

        addr = await Address("localhost", 8475, route)
        client = ToxiClient(addr)
        await client.start()

        # Create a new tunnel from toxiproxi to google.
        dest = await Address("www.google.com", 80, client.addr.route)
        tunnel = await client.new_tunnel(dest)
        assert(isinstance(tunnel, ToxiTunnel))

        downstream = ToxiToxic().downstream()
        toxic = downstream.add_latency(100)
        await tunnel.new_toxic(toxic)

        print(f"Tunnel serv port = {tunnel.port}")
        print(d_vals(server.tunnel_servs)[0].servers)
        s = d_vals(server.tunnel_servs)[0].servers[0]
        print(s)
        print(s[2].sock.getsockname())
        tun_pipe, tun_tup = await tunnel.get_pipe()
        print(tun_tup)
        await tun_pipe.send(b"test", tun_pipe)

        tunnel_serv = d_vals(server.tunnel_servs)[0]
        print(tunnel_serv.clients)
        # assert(len(tunnel_serv.clients) == 0)
        print("close tun pipe")
        await tun_pipe.close()
        await asyncio.sleep(3)
        #await server.close()
        print(tunnel_serv.clients)
        await asyncio.sleep(1)
        #print(s[2].tcp_server_task)

        
        #print(s)

        #
        #ask_exit()
"""

if __name__ == '__main__':
    pass



import multiprocessing
from p2pd.test_init import *
from p2pd.base_stream import *
from p2pd.turn_defs import *
from p2pd.turn_client import *
from p2pd.http_client_lib import *

async def get_turn_client(af, interface, turn_offset):
    # Get a route for this interface.
    route = await interface.route(af).bind()
    dest = await Address(
        TURN_SERVERS[turn_offset]["host"],
        TURN_SERVERS[turn_offset]["port"],
        route
    ).res()

    # Implement the TURN protocol for UDP send / recv.
    client = TURNClient(
        turn_addr=dest,
        turn_user=TURN_SERVERS[turn_offset]["user"],
        turn_pw=TURN_SERVERS[turn_offset]["pass"],
        turn_realm=TURN_SERVERS[turn_offset]["realm"],
        route=route
    )

    # Enable blank UDP headers.
    #client.toggle_blank_rudp_headers(True)

    # Wait for authentication and relay address allocation.
    await async_wrap_errors(
        client.start(),
        timeout=10
    )

    # Wait for the client to be ready.
    await client.client_tup_future
    await client.relay_tup_future
    return client

asyncio.set_event_loop_policy(SelectorEventPolicy())
class TestTurn(unittest.IsolatedAsyncioTestCase):
    async def test_turn_duel_ifs(self):
        # Offset of turn server to use.
        turn_offset = 0

        # Load interface list.
        netifaces = await init_p2pd()
        ifs, af = await duel_if_setup(netifaces)
        if af is None:
            return

        # Will store the turn clients.
        turn_clients = []
        relay_tup = None
        for interface in ifs:
            # Get turn client.
            client = await get_turn_client(
                af,
                interface,
                turn_offset
            )

            # Save to list of clients.
            turn_clients.append(client)

        # Each turn client white lists the others external IP.
        # Thereby allowing msgs from that interface through.
        for if_index in range(0, len(ifs)):
            # Sending interface and it's external IP
            interface = ifs[if_index]
            route = await interface.route(af).bind()
            peer_tup = (route.ext(), 0)

            # Accept the other interfaces external IP.
            turn_client = turn_clients[if_index - 1]
            turn_client.subscribe(SUB_ALL)
            await turn_client.accept_peer(peer_tup, peer_tup)

        # Test message receipt for both clients.
        msg = b"hello, world!"
        for if_index in range(0, len(ifs)):
            # Perspective is this if to that turn client.
            # The turn client is on another interface.
            interface = ifs[if_index]
            route = await interface.route(af).bind()
            turn_client = turn_clients[if_index - 1]

            # Send data to the relay endpoint.
            relay_tup = await turn_client.relay_tup_future
            for i in range(0, 3):
                await turn_clients[if_index].send(msg, relay_tup)

            # Recv data back.
            out = await turn_client.recv(SUB_ALL, 2)
            assert(msg in out)

        # Cleanup
        for turn_client in turn_clients:
            await turn_client.close()

    async def test_turn(self):
        #print(sys.modules.keys())
        # Network interface details.
        log(">>> test_turn")
        n = 0
        i = await Interface()
        af = i.supported()[0]
        r = await i.route(af).bind()

        # Address of a TURN server.
        dest = await Address(
            TURN_SERVERS[n]["host"],
            TURN_SERVERS[n]["port"],
            r
        ).res()

        # Implement the TURN protocol for UDP send / recv.
        client = await get_turn_client(
            af,
            i,
            n
        )

        # Disabled until valid net debug available.
        if 0:
            # Server hosting a PHP script that can send traffic.
            net_debug_url = NET_DEBUG_PROVIDERS[af][0]
            url_parts = await url_res(r, net_debug_url)
            outbound_ip = await url_open(
                r,
                url_parts,
                {
                    "action": "host",
                    "version": str(af_to_v(af))
                },
                timeout=6,
                throttle=1
            )

            print(outbound_ip)

            # Whitelist outbound addr of server.
            # Note that the send port doesn't matter here.
            bind_port = 0
            peer_tup = (outbound_ip, bind_port)
            print(peer_tup)
            await client.accept_peer(peer_tup, peer_tup)
            peer_tup = ("153.92.0.27", bind_port)
            await client.accept_peer(peer_tup, peer_tup)

            # Interested in any messages to queue.
            client.subscribe(SUB_ALL)

            #await client.accept_peer(client_tup, relay_tup)
            #await client.accept_peer(relay_tup, relay_tup)
            """
            m = TurnMessage(msg_type=TurnMessageMethod.SendResponse, msg_code=TurnMessageCode.Request)
            m.write_attr(
                TurnAttribute.Data,
                b"test message to send."
            )
            buf, _ = TurnMessage.unpack(m.encode())
            buf.write_credential(client.turn_user, client.realm, client.nonce)
            buf.write_hmac(client.key)
            """

            # Attempt to get a reply at our relay address.
            # Do it up to 3 times due to UDP being unreliable.
            got_reply = False
            for i in range(0, 3):
                http_route = copy.deepcopy(r)
                await http_route.bind()
                out = await url_open(
                    http_route,
                    url_parts,
                    {
                        "action": "hello",
                        "proto": "udp",
                        "host": relay_tup[0],
                        "port": str(relay_tup[1]),
                        "bind": str(bind_port)
                    },
                    timeout=6,
                    throttle=1
                )
                print(out)

                # Attempt to read the message back.
                out = await client.recv(SUB_ALL, 2)
                print(out)
                if out == b"hello":
                    got_reply = True
                    break

            # Make sure we got a valid reply.
            self.assertTrue(got_reply)
        
        # Test refresh.
        client.lifetime = 0
        await async_retry(
            lambda: client.refresh_allocation(), count=3, timeout=5
        )
        self.assertTrue(client.lifetime)

        # Cleanup the client.
        await client.close()

if __name__ == '__main__':
    main()


from .p2p_pipe import *
from .nat import *

TXT = {
    "p2p_strat": {
        P2P_DIRECT: "direct connect",
        P2P_REVERSE: "reverse connect",
        P2P_PUNCH: "tcp punch",
        P2P_RELAY: "turn relay"
    },
    "nat": {
        OPEN_INTERNET: "open internet (no NAT)",
        SYMMETRIC_UDP_FIREWALL: "possible firewall (no NAT)",
        FULL_CONE: "full cone",
        RESTRICT_NAT: "restrict reuse",
        RESTRICT_PORT_NAT: "restrict port",
        SYMMETRIC_NAT: "symetric",
        BLOCKED_NAT: "unknown (all responses blocked)"
    },
    "delta": {
        NA_DELTA: "not applicable",
        EQUAL_DELTA: "equal delta (local port == mapped port)",
        PRESERV_DELTA: "preserving delta ((local port + dist) == (mapped_start + dist))",
        INDEPENDENT_DELTA: "independent delta (rand port == (last_mapped += delta))",
        DEPENDENT_DELTA: "dependent delta ((local port += [1 to delta]) == (last_mapped += [1 to delta]))",
        RANDOM_DELTA: "random delta (local port == rand port)"
    }
}



from .net import *

"""
    IRC_SERVERS1 = [
        "irc.libera.chat",
        "irc.oftc.net",
        "irc.uni-erlangen.de",
        "irc.undernet.org",
        "irc.uworld.se",
        "irc.efnet.nl",
        "irc.freenode.net",
        "irc.quakenet.org",
        "irc.hackint.org",
        "irc.dal.net",
        "irc.chathispano.com",
        "irc.p2p-network.net",
        "irc.kampungchat.org",
        "irc.simosnap.com",
        "irc.hybridirc.com",
        "irc.explosionirc.net",
        "irc.gamesurge.net",
        "irc.snoonet.org",
        "irc.link-net.org",
        "irc.esper.net",
        "irc.abjects.net",
        "irc.synirc.net",
        "irc.europnet.org",
        "irc.konnectchatirc.net",
        "irc.oltreirc.eu",
        "irc.irchighway.net",
        "irc.irc-nerds.net",
        "irc.scenep2p.net",
        "irc.orixon.org",
        "irc.coders-irc.net",
        "irc.openjoke.org",
        "irc.digitalirc.org",
        "irc.darkscience.net",
        "irc.tilde.chat",
        "irc.geeknode.org",
        "irc.slashnet.org",
        "irc.sohbet.net",
        "irc.w3.org",
        "irc.IRCGate.it",
        "irc.chaat.fr",
        "irc.rootworld.net",
        "irc.darkfasel.net",
        "irc.freeunibg.eu",
        "irc.euirc.net",
        "irc.geekshed.net",
        "irc.globalirc.it",
        "irc.chatbg.info",
        "irc.furnet.org",
        "irc2.irccloud.com",
        "irc.xxxchatters.com",
        "irc.bol-chat.com",
        "irc.sorcery.net",
        "irc.dejatoons.net",
        "irc.technet.chat",
        "irc.kalbim.net",
        "irc.ptnet.org",
        "irc1.ptirc.org",
        "irc.skychatz.org",
        "irc.axon.pw",
        "irc.tamarou.com",
        "irc.mindforge.org",
        "irc.abandoned-irc.net",
        "irc.epiknet.org",
        "irc.allnetwork.org",
        "irc1.net-tchat.fr",
        "irc.swiftirc.net",
        "irc.afternet.org",
        "irc2.chattersnet.nl",
        "irc.allnightcafe.com",
        "irc.evilnet.org",
        "irc.rezosup.org",
        "irc.recycled-irc.net",
        "irc2.acc.umu.se",
        "irc.mibbit.net",
        "irc.pirc.pl",
        "irc.atrum.org",
        "irc.chatlounge.net",
        "irc.librairc.net",
        "irc.lunarirc.net",
        "irc.bigua.org",
        "irc.smurfnet.ch",
        "irc.ouch.chat",
        "irc.lewdchat.com",
        "irc.deutscher-chat.de",
        "irc.scuttled.net",
        "irc.chat.com.tr",
        "irc.spotchat.org",
        "irc.gigairc.net",
        "irc.darkworld.network",
        "irc.zenet.org",
        "irc.scoutlink.net",
        "irc.do-dear.com",
        "irc.luatic.net",
        "irc.roircop.info",
        "irc.forumcerdas.net",
        "irc.darkmyst.org"
    ]

    #IRC_SERVERS1 = ["irc.darkmyst.org"]

    # Taken from hexchat
    # Lets see what happens.
    IRC_SERVERS1 = [
        "pirc.pl",
        "newserver",
        "irc.2600.net",
        "global.acn.gr",
        "irc.afternet.org",
        "irc.data.lt",
        "irc.omicron.lt",
        "irc.vub.lt",
        "irc.anthrochat.net",
        "arcnet-irc.org",
        "irc.austnet.org",
        "irc.azzurra.org",
        "irc.canternet.org",
        "irc.chat4all.org",
        "irc.chatjunkies.org",
        "irc.unibg.net",
        "irc.chatpat.bg",
        "irc.chatspike.net",
        "irc.dairc.net",
        "us.dal.net",
        "irc.darkmyst.org",
        "irc.darkscience.net",
        "irc.drk.sc",
        "irc.darkscience.ws",
        "irc.d-t-net.de",
        "irc.digitalirc.org",
        "irc.dosers.net",
        "irc.choopa.net",
        "efnet.port80.se",
        "irc.underworld.no",
        "efnet.deic.eu",
        "irc.enterthegame.com",
        "irc.entropynet.net",
        "irc.esper.net",
        "irc.euirc.net",
        "irc.europnet.org",
        "irc.fdfnet.net",
        "irc.gamesurge.net",
        "irc.geekshed.net",
        "irc.german-elite.net",
        "irc.gimp.org",
        "irc.gnome.org",
        "irc.globalgamers.net",
        "irc.hackint.org",
        "irc.eu.hackint.org",
        "irc.hashmark.net",
        "irc.icq-chat.com",
        "irc.interlinked.me",
        "irc.irc4fun.net",
        "irc.irchighway.net",
        "open.ircnet.net",
        "irc.irctoo.net",
        "irc.kbfail.net",
        "irc.libera.chat",
        "irc.liberta.casa",
        "irc.librairc.net",
        "irc.link-net.org",
        "irc.mindforge.org",
        "irc.mixxnet.net",
        "irc.oceanius.com",
        "irc.oftc.net",
        "irc.othernet.org",
        "irc.oz.org",
        "irc.krstarica.com",
        "irc.pirc.pl",
        "irc.ptnet.org",
        "uevora.ptnet.org",
        "claranet.ptnet.org",
        "sonaquela.ptnet.org",
        "uc.ptnet.org",
        "ipg.ptnet.org",
        "irc.quakenet.org",
        "irc.rizon.net",
        "irc.tomsk.net",
        "irc.run.net",
        "irc.ru",
        "irc.lucky.net",
        "irc.serenity-irc.net",
        "irc.simosnap.com",
        "irc.slashnet.org",
        "irc.snoonet.org",
        "irc.sohbet.net",
        "irc.sorcery.net",
        "irc.spotchat.org",
        "irc.station51.net",
        "irc.stormbit.net",
        "irc.swiftirc.net",
        "irc.synirc.net",
        "irc.techtronix.net",
        "irc.tilde.chat",
        "irc.servx.org",
        "i.valware.uk",
        "irc.tripsit.me",
        "newirc.tripsit.me",
        "coconut.tripsit.me",
        "innsbruck.tripsit.me",
        "us.undernet.org",
        "irc.xertion.org"
    ]

    # These servers are taken from mirc.
    IRC_SERVERS1 = ['irc.dal.net', 'irc.efnet.org', 'open.ircnet.net', 'irc.libera.chat', 'irc.quakenet.org', 'irc.rizon.net', 'irc.snoonet.org', 'irc.swiftirc.net', 'irc.undernet.org', 'irc.scuttled.net', 'irc.abjects.net', 'irc.afternet.org', 'irc.data.lt', 'irc.allnetwork.org', 'irc.alphachat.net', 'irc.atrum.org', 'irc.austnet.org', 'irc.axon.pw', 'irc.ayochat.or.id', 'irc.azzurra.org', 'irc.beyondirc.net', 'irc.bolchat.com', 'ssl.bongster.de', 'irc.brasirc.com.br', 'irc.canternet.org', 'irc.chat4all.org', 'irc.chatspike.net', 'irc.chatzona.org', 'irc.cncirc.net', 'irc.coolsmile.net', 'irc.darenet.org', 'irc.d-t-net.de', 'irc.darkfasel.net', 'irc.darkmyst.org', 'irc.darkscience.net', 'irc.darkworld.network', 'irc.dejatoons.net', 'irc.desirenet.org', 'irc.ecnet.org', 'irc.epiknet.org', 'irc.esper.net', 'irc.euirc.net', 'irc.europnet.org', 'irc.evolu.net', 'irc.explosionirc.net', 'irc.fdfnet.net', 'irc.fef.net', 'irc.financialchat.com', 'irc.forestnet.org', 'irc.FreeUniBG.eu', 'irc.gamesurge.net', 'irc.geeknode.org', 'irc.geekshed.net', 'irc.german-elite.net', 'irc.gigairc.net', 'irc.gimp.org', 'irc.globalgamers.net', 'irc.goodchatting.com', 'irc.hackint.org', 'irc.hybridirc.com', 'irc.icq-chat.com', 'irc.immortal-anime.net', 'irc.indymedia.org', 'irc.irc-hispano.org', 'irc.irc2.hu', 'irc.irc4fun.net', 'irc.ircgate.it', 'irc.irchighway.net', 'irc.ircsource.net', 'irc.irctoo.net', 'irc.ircube.org', 'irc.ircworld.org', 'irc.irdsi.net', 'irc.kampungchat.org', 'irc.knightirc.net', 'irc.krey.net', 'irc.krono.net', 'irc.librairc.net', 'irc.lichtsnel.nl', 'irc.link-net.be', 'irc.luatic.net', 'irc.maddshark.net', 'irc.magicstar.net', 'irc.perl.org', 'irc.mibbit.net', 'irc.mindforge.org', 'irc.nationchat.org', 'irc.nightstar.net', 'irc.nullirc.net', 'irc.oftc.net', 'irc.oltreirc.net', 'irc.openjoke.org', 'irc.lt-tech.org', 'irc.orixon.org', 'irc.oz.org', 'irc.p2p-network.net', 'irc.phat-net.de', 'irc.krstarica.com', 'irc.pirc.pl', 'irc.ptnet.org', 'irc.recycled-irc.net', 'irc.retroit.org', 'irc.rezosup.org', 'irc.rusnet.org.ru', 'irc.scarynet.org', 'irc.serenity-irc.net', 'irc.shadowfire.org', 'irc.shadowworld.net', 'irc.simosnap.com', 'irc.SkyChatz.org', 'irc.skyrock.net', 'irc.slacknet.org', 'irc.slashnet.org', 'irc.smurfnet.ch', 'irc.sorcery.net', 'irc.spotchat.org', 'irc.st-city.net', 'irc.starlink-irc.org', 'irc.starlink.org', 'irc.staynet.org', 'irc.stormbit.net', 'irc.synirc.net', 'irc.technet.chat', 'irc.tilde.chat', 'irc.tweakers.net', 'irc.undermind.net', 'irc.wenet.ru', 'irc.whatnet.org', 'irc.wixchat.org', 'irc.worldirc.org', 'irc.xertion.org', 'irc.xevion.net']


    IRC_SERVERS1 = ['irc.oftc.net', 'irc.euirc.net', 'irc.xxxchatters.com', 'irc.swiftirc.net', 'irc.darkmyst.org', 'irc.chatjunkies.org', 'irc.dosers.net', 'irc.entropynet.net',  'irc.liberta.casa', 'irc.financialchat.com', 'irc.irc2.hu', 'irc.phat-net.de', 'irc.slacknet.org', 'irc.tweakers.net']
"""

irc_servers = [
    {
        'domain': 'irc.oftc.net',
        'afs': [IP4, IP6],

        # 20 jul 2002
        "creation": 1027087200
    },
    {
        'domain': 'irc.euirc.net',
        'afs': [IP4, IP6],


        # 19 sep 2000
        "creation": 969282000
    },
    {
        'domain': 'irc.xxxchatters.com',
        'afs': [IP4],

        # 9 march 2007
        'creation': 1173358800
    },
    {
        'domain': 'irc.swiftirc.net',
        'afs': [IP4, IP6],

        # 10 march 2007
        'creation': 1173445200
    },
    {
        'domain': 'irc.darkmyst.org',
        'afs': [IP4, IP6],

        # 26 nov 2002
        'creation': 1038229200
    },
    {
        'domain': 'irc.chatjunkies.org',
        'afs': [IP4],

        # 28 june 2002
        'creation': 1025186400
    },
    {
        'domain': 'irc.dosers.net',
        'afs': [IP4],

        # 20 may 2020
        'creation': 1590501600
    },
    {
        'domain': 'irc.entropynet.net',
        'afs': [IP4, IP6],

        # 11 sep 2011
        'creation': 1312984800
    },
    {
        'domain': 'irc.liberta.casa',
        'afs': [IP4, IP6],

        # 7 feb 2020
        'creation': 1580994000
    },
    {
        'domain': 'irc.financialchat.com',
        'afs': [IP4],

        # 1 aug 2002
        'creation': 1028210400
    },
    {
        'domain': 'irc.irc2.hu',
        'afs': [IP4],

        # 4 jan 2004
        'creation': 1073134800
    },
    {
        'domain': 'irc.phat-net.de',
        'afs': [IP4, IP6],

        # 6 nov 2000
        'creation': 975848400
    },
    {
        'domain': 'irc.slacknet.org',
        'afs': [IP4, IP6],

        # 20 aug 2000
        'creation': 966434400
    },
    {
        'domain': 'irc.tweakers.net',
        'afs': [IP4, IP6],

        # 30 apr 2002
        'creation': 1020088800
    },
]

# Sort servers by creation date (oldest first.)
irc_servers = sorted(irc_servers, key=lambda k: k['creation'])

# Create server lists by af.
IRC_SERVERS = {
    IP4: [s for s in irc_servers if IP4 in s['afs']],
    IP6: [s for s in irc_servers if IP6 in s['afs']],

}



from p2pd.test_init import *
from p2pd.address import *
from p2pd.net import *
from p2pd.interface import *
from p2pd.stun_client import *
import socket
class TestSock(unittest.IsolatedAsyncioTestCase):
    async def test_reuse_port(self):
        s1 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            s1.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except:
            pass
        s1.bind(('', 0))
        #s1.connect(("www.google.com", 80))

        port = s1.getsockname()[1]
        s2 = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            s2.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except:
            pass
        s2.bind(('', port))
        #s2.connect(("www.google.com", 80))

        s1.close()
        s2.close()

    async def test_socket_factory_connect(self):
        loop = asyncio.get_event_loop()
        i = await Interface().start_local()
        af = i.supported()[0]
        r = await i.route(af).bind(0)
        d = await Address("8.8.8.8", 53, r)
        s = await socket_factory(route=r, dest_addr=d, sock_type=TCP, conf=NET_CONF)
        con_task = asyncio.create_task(
            loop.sock_connect(
                s, 
                d.tup
            )
        )

        await asyncio.wait_for(con_task, 2)
        if s is not None:
            s.close()

    async def test_high_port_reuse(self):
        # Config for reuse.
        conf = copy.deepcopy(NET_CONF)
        conf["reuse_addr"] = True

        # Load default interface.
        i = await Interface().start_local()
        r = i.route()

        # Make a new socket bound to a high order port.
        high_sock, high_port = await get_high_port_socket(r)

        # Make a new socket that shares the same port.
        r = await i.route().bind(high_port)
        reuse_sock = await socket_factory(r, conf=conf)

        # Cleanup both socket handles.
        high_sock.close()
        reuse_sock.close()



if __name__ == '__main__':
    main()

"""
one of the nic ips is not working. why would this break the
stun code though?
"""


from p2pd import *

# Warning: this is very slow to start.
async def example():
    # Start our main node server.
    # The node implements your protocol.
    node = await start_p2p_node(
        # Set to true for port forwarding + pin holes.
        enable_upnp=False,
        #
        # Make sure node server uses different port
        # to other examples.
        port=NODE_PORT + 50 + 12
    )
    #
    # Strategies used to make a P2P connection.
    # Note that P2P_RELAY enables TURN.
    # (Coturn doesn't support self relay so removed.)
    strategies = [ P2P_DIRECT, P2P_REVERSE, P2P_PUNCH ]
    #
    """
    Spawns a new pipe from a P2P connection.
    In this case it's connecting to our own node server.
    There will be no barriers to do this so this will just use
    a plain direct TCP connection / P2P_DIRECT.
    Feel free to experiment with how it works.
    """
    pipe, success_type = await node.connect(node.addr_bytes, strategies)
    #
    # Do some stuff on the pipe ...
    # Cleanup.
    await pipe.close()
    await node.close()

if __name__ == '__main__':
    async_test(example)


import ctypes
import shlex
import asyncio
import tempfile
import os
import uuid
import platform
import sys
import subprocess
import base64
from .utils import *

"""
Windows doesn't always use UTF-8 for string encoding.
Depending on which functions you're using.
This completely screws encoding and decoding bytes / strs.
"""
os.environ["PYTHONIOENCODING"] = "utf-8"

"""
This requires the process to have been run as administrator
or with a UAC prompt. However, by using the technique of
writing self-contained Python files to disk that accomplish
privileged operations and then restart themselves as admin
it becomes trivial to do the whole process at run time.

E.g. use run_py_script with a script that calls
pyuac.runAsAdmin() if it's not an admin and is designed
to complete the privileged call - all as an external,
new, python process, while the original continues to
run with its original privileges.

My function nt_pshell already checks if powershell can
run scripts and if not -- automatically unrestricts
powershell and continues running the script.
"""
async def nt_set_pshell_unrestricted():
    import winreg

    def mk_unrestricted(path):
        # Open path to key for 'writing.'
        key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            path,
            access=winreg.KEY_SET_VALUE
        )

        # Set the sub key.
        # Create it if it doesn't exist.
        winreg.SetValueEx(
            key,
            "ExecutionPolicy",
            0,
            winreg.REG_SZ,
            "Unrestricted"
        )

        # We want to be certain the registry
        # has been updated before proceeding.
        winreg.FlushKey(key)

        # Cleanup.
        winreg.CloseKey(key)

    base_path = "SOFTWARE\\Microsoft\\PowerShell\\1\\ShellIds"
    pshell_path = base_path + "\\Microsoft.PowerShell"
    nostics_path = base_path + "\\ScriptedDiagnostics"
    mk_unrestricted(nostics_path)
    mk_unrestricted(pshell_path)

def nt_is_admin():
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except:
        return False

# Surrounds with DOUBLE quotes.
def mac_arg_escape(arg):
    black_list = "\"\\"
    buf = ""
    for ch in arg:
        if ch in black_list:
            buf += "\\" + ch
        else:
            buf += ch

    return '"' + buf + '"'

"""
Note: this function escapes an argument string
for Unix shell but surrounds it by single quotes.
It returns a single quoted string with the result.
Hence the surrounding quotes APPEAR escaped when
printing them.
"""
# Surrounds with SINGLE quotes.
def nix_arg_escape(arg):
    return shlex.quote(arg)

def win_arg_escape(arg, allow_vars=0):
    # Double all the backslashes before a double quote.
    # Then ensure the double quote is escaped.
    # Doubling up neutralizes double quotes that may be unbalanced.
    # Since this is done you ensure quotes are all escaped properly.
    arg = re.sub(r'(\\*)"', r'\1\1\\"', arg)

    # Double up continuous backslashes at the end of the string.
    arg = re.sub(r'(\\*)$', r'\1\1', arg)

    # Replace special characters.
    arg = re.sub('([()%!^<>&|;,])', r'^\1', arg)

    # Surround entire arg with quotes.
    # This avoids spaces breaking a command.
    arg = '"%s"' % (arg)

    return arg

"""
There is an issue with create_subprocess_shell on Windows 10
with the latest Python versions. When you try use this code
a window will pop up asking you how you want to open this
file. A hack I've found that works is to pass the command
you want to execute to powershell.

Example: 'powershell "route print"'
"""
async def cmd(value, io=None, timeout=10):
    # Output type.
    out_type = value
    null_out = to_type('', out_type)

    # Setup STDIN pipes.
    in_val = None
    if io is not None:
        in_val = asyncio.subprocess.PIPE

    this_dir = os.path.dirname(__file__)
    try:
        proc = await asyncio.create_subprocess_shell(
            value,
            stdin=in_val,
            stdout=asyncio.subprocess.PIPE,
            # stderr=asyncio.subprocess.PIPE,
            cwd=this_dir,
            shell=True
        )

        if timeout:
            try:
                task = proc.communicate(input=io)
                stdout, stderr = await asyncio.wait_for(task, timeout)
            except asyncio.TimeoutError:
                log(f"command {value} timed out")
                return null_out
        else:
            stdout, stderr = await proc.communicate(input=io)
    except NotImplementedError:
        # Backup function for old Python versions.
        @run_in_executor
        def blocking_cmd():
            try:
                proc = subprocess.run(value, stdout=subprocess.PIPE, shell=True, timeout=timeout)
                stdout = proc.stdout
                stderr = proc.stderr
            except subprocess.TimeoutExpired:
                log(f"command {value} timed out")
                return null_out

            return stdout, stderr
        
        # Use an executor to run the blocking command so the event loop isn't wrekt.
        stdout, stderr = await blocking_cmd()

    # Log any visible errors.
    if stderr is not None and len(stderr):
        log(f"cmd {value} stderr = {stderr}")

    # Return command output.
    if stdout is None:
        return null_out
    else:
        return to_type(stdout, out_type)
    
def powershell_encoded_cmd(ps1):
    unicode_bytes = ps1.encode("utf-16le")
    param = base64.b64encode(unicode_bytes)
    return to_s(param)

async def ps1_exec_trick(ps1):
    param = powershell_encoded_cmd(ps1)
    out = await cmd(f"powershell -encodedCommand {param}", timeout=None)
    return out

async def is_pshell_restricted():
    out = await cmd("powershell Get-ExecutionPolicy", timeout=None)
    return not "Unrestricted" in out

async def nt_pshell(value, timeout=10, is_unrestricted=None):
    # Allow powershell scripts to be run
    # by modifying registry if needed.
    if is_unrestricted is None:
        if(await is_pshell_restricted()):
            await nt_set_pshell_unrestricted()

    # Write a temp file into the temp dir
    # with the script to execute.
    tmp_dir = tempfile.gettempdir()
    cmd_path = os.path.join(tmp_dir, str(uuid.uuid4()) + ".ps1")
    with open(cmd_path, 'w') as f:
        f.write(value)
        f.flush()
        f.close()

    # Command to tell powershell to run the script.
    cmd_str = 'powershell.exe -file %s' % (
        win_arg_escape(cmd_path)
    )

    # Wait for powershell to execute the script.
    if timeout:
        out = await cmd(cmd_str, timeout=timeout)
    else:
        out = await cmd(cmd_str, timeout=None)

    # Delete the old script file.
    if os.path.exists(cmd_path):
        os.remove(cmd_path)

    return out

def get_arg_escape_func():
    if platform.system() == "Linux":
        return nix_arg_escape

    if platform.system() == "Windows":
        return win_arg_escape

    if platform.system() in ["Darwin", "FreeBSD"]:
        return mac_arg_escape

    return None

async def run_py_script(script, root_pw=None, cleanup=False):
    # Write a temp file into the temp dir
    # with the script to execute.
    out = None
    tmp_dir = tempfile.gettempdir()
    cmd_path = os.path.join(tmp_dir, str(uuid.uuid4()) + ".py")
    escape_arg = get_arg_escape_func()
    with open(cmd_path, 'w') as f:
        f.write(script)
        f.flush()
        f.close()

        # Sudo prefix.
        prefix = ""
        if platform.system() != "Windows":
            if root_pw != None:
                prefix = "{ echo %s; } | sudo -k -S " % (
                    escape_arg(root_pw)
                )

        # Build cmd to execute the new script.
        exe_cmd = '%s%s %s' % (
            prefix,
            escape_arg(sys.executable),
            escape_arg(cmd_path)
        )
        
        # Execute the command.
        out = await cmd(exe_cmd, timeout=None)

        # Delete the script file.
        if cleanup:
            if os.path.exists(cmd_path):
                os.remove(cmd_path)

    return out

def is_root():
    if platform.system() == "Windows":
        import ctypes
        if ctypes.windll.shell32.IsUserAnAdmin() == 0:
            return False
    else:
        if os.geteuid() != 0:
            return False

    return True

def win_uac():
    ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, " ".join(sys.argv), None, 1)

def ensure_root():
    if not is_root():
        raise Exception("root required for this code.")


from p2pd import *

async def msg_cb(msg, client_tup, pipe):
    await pipe.send(msg, client_tup)

async def example():
    #
    # Start default interface and get the first route.
    # No AF for route use i.supported()[0] 
    i = await Interface().start()
    route = await i.route().bind() # Port 0 = any unused port.
    #
    # Start the server and use msg_cb to process messages.
    server = await pipe_open(TCP, route, msg_cb=msg_cb)
    #
    # Connect to the server.
    # Use the IP of the route and unused port for the destination.
    dest = await Address(*server.sock.getsockname()[0:2], route).res()
    client = await pipe_open(TCP, route, dest)
    #
    # Send data to the server and check receipt.
    msg = b"test msg."
    await client.send(msg)
    out = await client.recv()
    assert(msg == out)
    #
    # Close both.
    await client.close()
    await server.close()

# From inside the async REPL.
if __name__ == '__main__':
    async_test(example)


from p2pd.test_init import *
import platform
import multiprocessing
from p2pd.net import VALID_AFS
from p2pd.cmd_tools import *
from p2pd import Interface, IP4, IP6
from p2pd.route_table import is_internet_if, get_route_table

RELATED_PLATFORMS = ["Linux", "Darwin", "FreeBSD"]

class TestRouteTable(unittest.IsolatedAsyncioTestCase):
    async def test_route_table(self):
        if platform.system() in RELATED_PLATFORMS:
            netifaces = await init_p2pd()
        
        applies = False
        one_worked = False
        for af in VALID_AFS:
            if platform.system() in RELATED_PLATFORMS:
                applies = True
                try:
                    table = await get_route_table(af)
                    if not len(table):
                        continue

                    # Get a default interface.
                    i = Interface(netifaces=netifaces)

                    # Check that the func to test internet interfaces works.
                    if i.nic_no:
                        # Bug here on Windows but Windows doesn't depend on this logic.
                        r = await is_internet_if(i.nic_no)
                    else:
                        r = await is_internet_if(i.name)
                        
                    one_worked = True
                except Exception:
                    log_exception()
                    pass

        if applies:
            if not one_worked:
                print("The route table code failed.")
                print("If the system has a route command this may be an error.")


if __name__ == '__main__':  
    main()


from p2pd import *

async def example():
    i = await Interface().start()
    nat = await i.load_nat()
    #
    # Test echo server with AF.
    stun_client = STUNClient(i, i.supported()[0])
    wan_ip = await stun_client.get_wan_ip()
    results = await stun_client.get_mapping(TCP)

if __name__ == '__main__':
    async_test(example)


from p2pd import *

class NetInfoServer(Daemon):
    async def msg_cb(self, msg, client_tup, pipe):
        # Parse HTTP message and handle CORS.
        req = await rest_service(msg, client_tup, pipe)
        #
        async def get_response():
            # Show information of the connection to the peer.
            p = req.api("/mapping")
            if p:
                sock = pipe.sock
                return {
                    "error": 0,
                    "fd": sock.fileno(),
                    "laddr": sock.getsockname(),
                    "raddr": sock.getpeername(),
                    "route": pipe.route.to_dict(),
                    "if": {
                        "name": pipe.route.interface.name
                    }
                }
            #
            # Fallback and serve all other purposes.
            return {
                "error": 2,
                "msg": "method not implemented"
            }
        #
        resp = await get_response()
        if resp is not None:
            await send_json(
                resp,
                req,
                client_tup,
                pipe
            )

async def example():
    # Default interface of your machine.
    # netifaces.interfaces() for names
    # or await load_interfaces() for a started list.
    i = await Interface().start()
    #
    # Server object inherits from a standard Daemon.
    server = NetInfoServer()
    #
    # Makes a Route aware of all Routes used for the server.
    # Might do this automatically in the future.
    server.set_rp(i.rp)
    #
    # Defines addresses and protocols to listen on.
    # Feel free to switch this up.
    await server.listen_all(
        # Listen on all routes for IP4 and IPv6.
        [i.rp[IP4], i.rp[IP6]],
        #
        # Port(s) to listen on.
        [20000],
        #
        # Interested in TCP and UDP.
        [TCP, UDP]
    )
    # Do a while sleep loop ...
    # Instead we'll just close.
    await server.close()

if __name__ == '__main__':
    async_test(example)


import random
from .address import *

STUN_PORT = 3478
MAX_MAP_NO = 100

# NAT TYPES ---------------------------------------------

# No NAT at all.
OPEN_INTERNET = 1

# There is no NAT but there is some kind of firewall.
SYMMETRIC_UDP_FIREWALL = 2

# Mappings are made for local endpoints.
# Then any destination can use the mapping to reach the local enpoint.
# Note: May be incorrectly detected if using TCP.
FULL_CONE = 3

# NAT reuses mapping if same src ip and port is used.
# Destination must be white listed. It can use any port to send replies on.
# Endpoint-independent
# Note: May be incorrectly detected if using TCP.
RESTRICT_NAT = 4

# Mappings reused based on src ip and port.
# Destination must be white listed and use the port requested by recipient.
# Endpoint-independent (with some limitations.)
# Note: May be incorrectly detected if using TCP.
RESTRICT_PORT_NAT = 5

# Different mapping based on outgoing hosts.
# Even if same source IP and port reused.
# AKA: End-point dependent mapping.
SYMMETRIC_NAT = 6

# No response at all.
BLOCKED_NAT = 7
# ---------------------------------------------------------

# DELTA types: ------------------
# Mappings are easy to reuse and reach.
NA_DELTA = 1 # Not applicable.
EQUAL_DELTA = 2
PRESERV_DELTA = 3 # Or not applicable like in open internet.
INDEPENDENT_DELTA = 4
DEPENDENT_DELTA = 5
RANDOM_DELTA = 6
# -------------------------------

EASY_NATS = [OPEN_INTERNET, FULL_CONE]
DELTA_N = [
    # Remote = Local.
    EQUAL_DELTA,

    # Remote x - y = local x - y
    PRESERV_DELTA,

    # Remote x; y = remote x + delta, local = anything
    INDEPENDENT_DELTA,

    # Remote x; y = remote x + delta only when local x + delta
    DEPENDENT_DELTA
]

# The NATs here have various properties that allow their
# mappings to be predictable under certain conditions.
PREDICTABLE_NATS = [
    # Once open - anyone can use the mapping.
    FULL_CONE,

    # Same local IP + port = same mapping.
    RESTRICT_NAT,
    
    # Same as above but reply port needs to match original dest.
    RESTRICT_PORT_NAT,
]

# NAT types that require specific reply ports.
FUSSY_NATS = [
    RESTRICT_PORT_NAT,
]

# Peer will be unreachable.
BLOCKING_NATS = [
    BLOCKED_NAT
]

# Punch modes.
TCP_PUNCH_LAN = 1
TCP_PUNCH_REMOTE = 2
TCP_PUNCH_SELF = 3

# Convenience funcs.
# delta, nat_type
f_is_open = lambda n, d: n in [OPEN_INTERNET, SYMMETRIC_UDP_FIREWALL]
f_can_predict = lambda n, d: n in PREDICTABLE_NATS
def f_is_hard(n, d):
    not_easy = n not in EASY_NATS
    return not_easy and d["type"] not in [PRESERV_DELTA, EQUAL_DELTA]

def delta_info(delta_type, delta_value):
    return {
        "type": delta_type,
        "value": delta_value
    }

def nat_info(nat_type=None, delta=None, map_range=None):
    # Defaults.
    delta = delta or delta_info(RANDOM_DELTA, 0)
    map_range = map_range or [1, MAX_PORT]
    nat_type = nat_type or RESTRICT_PORT_NAT

    # Main NAT dic with simple lookup types.
    nat = {
        "type": nat_type,
        "delta": delta, # type, value
        "range": map_range, # start, stop
        "is_open": f_is_open(nat_type, delta),
        "can_predict": f_can_predict(nat_type, delta),
        "is_hard": f_is_hard(nat_type, delta)
    }

    # Setup whether this NAT type is good for concurrent punches.
    bad_delta = delta["type"] in [INDEPENDENT_DELTA, DEPENDENT_DELTA]
    nat["is_concurrent"] = nat_type in EASY_NATS or not bad_delta

    # Return results.
    return nat

def valid_mappings_len(mappings):
    if not len(mappings):
        return 0

    if len(mappings) > MAX_MAP_NO:
        return 0

    return 1

"""
The algorithm for selecting the same connection is based on
choosing the highest remote port on the 'master.'
Thus, if multiple cons have the same remote
port (on port over-loaded NATs - this is possible) --
then the code will fail. What we want is unique mappings.
"""
def rmaps_strip_duplicates(rmaps):
    remote_list = []
    reply_list = []
    local_list = []
    filtered_list = []
    for rmap in rmaps:
        remote, reply, local = rmap
        if remote in remote_list:
            continue

        if reply in reply_list:
            continue

        if local in local_list:
            continue
        
        remote_list.append(remote)
        reply_list.append(reply)
        local_list.append(local)
        filtered_list.append(rmap)

    return filtered_list

def is_valid_rmaps(rmaps):
    if type(rmaps) != list:
        return 0

    for e in rmaps:
        if type(e) != list:
            return 0

        if len(e) != 3:
            return 0

        for port in e:
            if type(port) != int:
                return 0

        # remote.
        if not valid_port(e[0]):
            return 0

        # local.
        if not valid_port(e[2]):
            return 0

        # check reply.
        if e[1]:
            if not valid_port(e[1]):
                return 0

    return 1

"""
Determine any numerical patterns in port mapping allocations.

Symmetric NATs use a new mapping for each destination + port.
By default this means you can't predict mappings. However,
it may still be possible if the NAT uses a predictable value
between mappings -- hence this is checked for.

The code here also distinguishes between 'dependent' and
'independent' delta types. Independent delta means NAT
mappings increase by n no matter what the source port
might be. It may be useful to differentiate this from a
delta type that is dependent on local ports because it's
more likely to have collisions seeing as how any outgoing
connection will increase the mapping counter.

Therefore -- this NAT type is poorly suited to concurrent
TCP hole punching and worth making the distinction for
even though the port prediction code for both is identical.

Note: Some delta type detections dependent on comparing the
value of the previous result. If concurrency is enabled
they results will be 'out of order' or unpredictable and
lead to invalid results. Concurrency is thus disabled.
This code is very badly written but it's at least tested.
"""
async def delta_test(stun_client, test_no=8, threshold=5, proto=DGRAM, group="map", concurrency=True):
    """
    - When getting a list of ports to use for tests
    calculate the port to start at.
    - If another start port has been used in other tests
    make sure not to choose a starting port that conflicts
    with a port in another tests range.
    - port_dist = distance between successive ports
    from start_port to test_no, incremented by port_dist.
    """
    def get_start_port(port_dist, range_info=[]):
        # Random port skipping reserved and max ports.
        rand_start_port = lambda: random.randrange(
            4000, MAX_PORT - (test_no * port_dist)
        )

        # Return if no other port range to check for conflicts.
        if not range_info:
            return rand_start_port()
        else:
            new_start_port = None
            do_retry = 1
            while do_retry:
                # Try a rand port as the starting port.
                do_retry = 0
                new_start_port = rand_start_port()
                for other_range in range_info:
                    # Range is other_start_port to other_end_port inclusive.
                    other_dist, other_start_port = other_range
                    other_end_port = other_start_port + (test_no * other_dist)

                    # If it's in the same range as other_start_port retry.
                    lower_bound = new_start_port >= other_start_port
                    upper_bound = new_start_port <= other_end_port
                    if lower_bound and upper_bound:
                        do_retry = 1
                        break

            return new_start_port

    # Create a list of tasks to get a mapping for a port range.
    def get_port_tests(start_port, port_dist=1):
        # Single concurrent STUN req.
        conf = dict_merge(NET_CONF, {
            "packet_retry": 1,
            "retry_no": 1,
            "recv_timeout": 0.5,
            "addr_retry": 1,
            "dns_timeout": 0.5
        })

        # Return task list for tests.
        tasks = []
        for i in range(0, test_no):
            # If start is defined then calculate a list of ports.
            # Otherwise the OS assigns an unused port.
            if start_port:
                src_port = start_port + (i * port_dist)
            else:
                src_port = random.randrange(4000, MAX_PORT)

            # Get the mapping using STUN.
            async def result_wrapper(src_port):
                # Make sure port isn't in the reserved range.
                if src_port < 4000 and src_port != 0:
                    raise Exception("src less than 4k in mapping behavior.")
                
                # Avoid relying on any one server. 
                server = stun_client.rand_server()
                
                # Get mapping using specific source port.
                _, s, local, mapped, _, _ = await stun_client.get_mapping(
                    proto,
                    do_close=0,
                    source_port=src_port,
                    group=group,
                    conf=conf,
                    servers=[server]
                )

                # Return mapping results.
                return [local, mapped, s]

            # Allow for tests to be done concurrently.
            tasks.append(
                async_wrap_errors(
                    asyncio.wait_for(
                        result_wrapper(src_port),
                        2
                    )
                )
            )

        return tasks

    def get_delta_value(delta_no, dist_no, local_dist, preserv_dist, results):
        for i in range(0, len(results)):
            # Unpack result.
            local, mapped, s = results[i]
            socks.append(s)
            if mapped is None:
                continue

            # Set previous result if available.
            prev_result = None
            if i != 0:
                if results[i - 1][MAPPED_INDEX] is not None:
                    prev_result = results[i - 1]

            # Preserving NAT.
            if local == mapped:
                delta_no[EQUAL_DELTA] = delta_no.get(EQUAL_DELTA, 0) + 1

            # Comparison tests.
            if prev_result is not None:
                # Skip invalid results.
                prev_local = prev_result[LOCAL_INDEX]
                prev_mapped = prev_result[MAPPED_INDEX]
                if not prev_local or not prev_mapped:
                    continue

                # Preserving delta if true.
                _local_dist = abs(field_dist(local, prev_local, MAX_PORT))
                mapped_dist = abs(field_dist(mapped, prev_mapped, MAX_PORT))
                if mapped_dist == _local_dist:
                    # Otherwise its preserving.
                    if mapped != local:
                        preserv_dist[mapped_dist] = 1
                        delta_no[PRESERV_DELTA] = delta_no.get(PRESERV_DELTA, 0) + 1
                else:
                    # Delta mapping dist.
                    # Plus one NAT type now here.
                    if _local_dist != 1:
                        dist_no[mapped_dist] = dist_no.get(mapped_dist, 0) + 1
                    else:
                        local_dist[mapped_dist] = local_dist.get(mapped_dist, 0) + 1

    # Offset names for port test results.
    LOCAL_INDEX = 0 # Source port.
    MAPPED_INDEX = 1 # External mapped port.
    SOCK_INDEX = 2 # Ref to sock used for STUN test.
    socks = []

    # Used for info about port ranges used for tests.
    # [ [ dist, start_port ], ... ]
    range_info = []

    # Do first port tests with random local ports.
    tasks = get_port_tests(0)
    if concurrency:
        results = await asyncio.gather(*tasks)
    else:
        results = []
        for task in tasks:
            result = await task
            results.append(result)

    """
    Check for:
        equal delta: src_port == mapped_port
        preserving delta NATs: dist(src_a, src_b) == dist(map_a, map_b)
        delta n (independent): dist(map_a, map_b) == delta n
    """ 
    delta_no = {}
    dist_no = {}
    preserv_dist = {}
    local_dist = {}

    # Close previous sockets.
    get_delta_value(delta_no, dist_no, local_dist, preserv_dist, results)
    socks = await sock_close_all(socks)

    # See if any of the above tests succeeded.
    test_names = [ EQUAL_DELTA, PRESERV_DELTA ]
    if len(preserv_dist) <= 1:
        test_names = [ EQUAL_DELTA]

    for test_name in test_names:
        if test_name not in list(delta_no.keys()):
            continue

        no = delta_no[test_name]
        if no >= threshold:
            return delta_info( test_name, 0 )
    for port_dist in list(dist_no.keys()):
        no = dist_no[port_dist]
        if no >= threshold:
            return delta_info( INDEPENDENT_DELTA, port_dist )

    """
    Check for:
        dependent delta check = also requires delta increase in src port
        delta n (dependent): dist(map_a, map_b) == local src++
    """
    delta_no = {}
    dist_no = {}
    preserv_dist = {}
    local_dist = {}

    # Get mapping results for fixed delta.
    start_port = get_start_port(1, range_info)
    tasks = get_port_tests(start_port)
    if concurrency:
        results = await asyncio.gather(*tasks)
    else:
        results = []
        for task in tasks:
            result = await task
            results.append(result)

    get_delta_value(delta_no, dist_no, local_dist, preserv_dist, results)

    # Check for deltas that satisfy success threshold.
    for port_dist in list(local_dist.keys()):
        no = local_dist[port_dist]
        if no >= threshold:
            return delta_info( DEPENDENT_DELTA, port_dist )

    # Return delta value.
    return delta_info( RANDOM_DELTA, 0 )

"""
A NAT may only allocate mappings from within a fixed range.
Starlink's default router is an example of this.
It has an allocation range of like 35000 - MAX_PORT.
It probably chose to do this for security reasons.
E.g. most known services use low range ports. If it only
uses higher range ports for mappings it won't show up
for port scanners as much.
""" 
def nats_intersect_range(our_nat, their_nat, test_no):
    # Calculate intersection range.
    is_intersect = range_intersects(our_nat["range"], their_nat["range"])
    if is_intersect:
        # A range that represents an overlapping portion.
        # Between our[range] and their[range] if any.
        r = intersect_range(
            our_nat["range"],
            their_nat["range"]
        )

        # If range is long enough then use it.
        range_no = r[1] - r[0]
        if range_no >= test_no:
            use_range = r
        else:
            use_range = our_nat["range"]
    else:
        use_range = our_nat["range"]

    # Ensure bind ports don't end up in low port ranges.
    if use_range[0] < 1024:
        use_range[0] = 1024
        if use_range[0] >= use_range[1]:
            raise Exception("Can't find intersecting port range.")
        
    return use_range

###############################################################
# [ [ local, remote, reply, sock ] ... ]
async def get_single_mapping(mode, rmap, last_mapped, use_range, our_nat, stun_client):
    # Allow last mapped to be modified from inside func.
    last_local, last_remote = last_mapped

    # Need to bind to a specific port.
    remote_port, reply_port, their_local = rmap
    bind_port = reply_port or remote_port

    """
    Normally the code tries to use the same port as
    the recipient to simplify the code and make things
    more resilient. But if you're trying to punch yourself
    using the same local port will be impossible. Choose
    a non-conflicting port.
    """
    if mode == TCP_PUNCH_SELF:
        for _ in range(0, 1000):
            bind_port = random.randrange(2000, MAX_PORT)
            if bind_port != remote_port:
                break

    # If we're port restricted specify we're happy to use their mapping.
    # This may not be possible if our delta is random though.
    our_reply = bind_port if our_nat["type"] == RESTRICT_PORT_NAT else 0

    # Use their mapping as-is.
    if our_nat["is_open"]:
        return [[bind_port, bind_port, 0, None], last_mapped]

    # If preserving try use their mapping.
    if our_nat["delta"]["type"] == EQUAL_DELTA:
        if not in_range(bind_port, our_nat["range"]):
            bind_port = from_range(use_range)

        return [[bind_port, bind_port, 0, None], last_mapped]

    # NAT preserves distance between local ports in remote ports.
    if our_nat["delta"]["type"] == PRESERV_DELTA:
        # Try use their port but make sure it fits in our range.
        if not in_range(bind_port, our_nat["range"]):
            bind_port = from_range(use_range)

        # How far away is our last mapping from desired port.
        # Delta dist will wrap inside any assumed range.
        dist = abs(n_dist(last_remote, bind_port))
        next_local = port_wrap(last_local + dist)

        # Return results.
        return [[next_local, bind_port, our_reply, None], last_mapped]

    """
    The routers NATs allocate mappings from a known range as
    determined by doing an inital large number of STUN tests.
    This means that when near the end of a range an
    'independent' or 'dependent' type NAT will wrap around
    back to the start of the range.

    The problem with this is if the ranges provided to the
    function aren't precise then when the ports 'wrap around'
    they will land on ports before the provided range.
    This means that the predicted port will be wrong and
    the hole punching will fail.

    The ranges need to be precise should they not be
    from 1 - MAX_PORT and use these type of deltas.
    Increasing the test_no and rounding to powers of 2
    may help to increase accuracy of the ranges.
    """
        
    # Poor concurrency support.
    if our_nat["delta"]["type"] == INDEPENDENT_DELTA:
        # We can use anything for a local port.
        # The remote mappings have a pattern regardless of local tuples.
        next_local = from_range([2000, MAX_PORT])
        next_remote = field_wrap(
            last_remote + our_nat["delta"]["value"],
            use_range
        )

        # Return port predictions.
        # These allocations apply even if strict port NAT.
        # But we tell other side to use a specific mapping for coordination.
        last_mapped = [next_local, next_remote]
        return [[next_local, next_remote, our_reply, None], last_mapped]

    # Poor concurrency support.
    if our_nat["delta"]["type"] == DEPENDENT_DELTA:
        next_local = port_wrap(last_local + 1)
        next_remote = field_wrap(
            last_remote + our_nat["delta"]["value"],
            use_range
        )

        # Return port predictions.
        # These allocations apply even if strict port NAT.
        # But we tell other side to use a specific mapping for coordination.
        last_mapped = [next_local, next_remote]
        return [[next_local, next_remote, our_reply, None], last_mapped]

    # Delta type is random -- get a mapping from STUN to reuse.
    # If we're port restricted then set our reply port to the STUN port.
    if our_nat["type"] in PREDICTABLE_NATS:
        # Get a mapping to use.
        _, s, local_port, remote_port, _, _ = await stun_client.get_mapping(
            proto=STREAM
        )

        # Calculate reply port.
        our_reply = 3478 if our_nat["type"] == RESTRICT_PORT_NAT else 0
        # TODO: Could connect to STUN port in their range.

        # Return results.
        return [[local_port, remote_port, our_reply, s], last_mapped]

    """
    The hardest NATs to support are 'symmetric' type NATs that only
    allow mappings to be used per [src ip, src port, dest ip, dest port].
    The only way to support symmetric NATs is if they also have
    a non-random delta. If this point is reached then they are likely
    are heavily restricted NAT with a random delta.
    """

    raise Exception("Can't predict this NAT type.")

"""
The code in this function is mostly error checking for
'restricted port NATs.' It checks that its possible
to predict mappings based on the NAT types in play.
"""
def nats_can_predict(our_nat, their_nat):
    ###############################################################
    # Increase test no if its a restricted port NAT.
    our_strict = our_nat["type"] == RESTRICT_PORT_NAT
    other_strict = their_nat["type"] == RESTRICT_PORT_NAT
    if our_strict and other_strict:
        # Error scenario for two strict port NATs with rand deltas.
        our_rand_delta = our_nat["delta"]["type"] == RANDOM_DELTA
        their_rand_delta = their_nat["delta"]["type"] == RANDOM_DELTA
        if our_rand_delta or their_rand_delta:
            raise Exception("Two strict port nats need non-rand deltas.")

    # If either side is port restrict make sure its partner can satisfy reply port.
    use_stun_port = 0
    if our_strict or other_strict:
        for nats in [[our_nat, their_nat], [their_nat, our_nat]]:
            # Switch which side is restricted.
            # Both may be restricted -- in which case both need non-rand delta.
            strict, unrestrict = nats
            if unrestrict["type"] == RESTRICT_PORT_NAT:
                strict, unrestrict = unrestrict, strict

            # If strict side has rand delta and partner cant satisfy reply port.
            # There are multiple ways to satisfy the reply port which are checked.
            # Raise an error condition if certain failure.
            strict_rand_delta = strict["delta"]["type"] == RANDOM_DELTA
            if strict_rand_delta and unrestrict["is_hard"]:
                raise Exception("Unable to satisfy mapping for strict port type.")

            # The reply port here is the STUN port as they use STUN to get a mapping.
            # Unable to satisfy reply port due to allocation range.
            if strict_rand_delta and not in_range(STUN_PORT, unrestrict["range"]):
                raise Exception("Can't support reply port 3478 for strict NAT.")

            # Use STUN port when we generate the fake 'their_mappings.'
            """
            OS' usually require admin or root to listen to the first 1024 ports.
            It is quite convenient that STUN listens on port 3478 or some
            clients behind certain NATs would need to be run as root to
            match the reply port of its peer.
            """
            if strict_rand_delta:
                use_stun_port = 1

    return use_stun_port

# Generates a list of local ports to remote mapped ports,
# and / or required reply ports for TCP hole punching.
async def get_nat_predictions(mode, stun_client, our_nat, their_nat, their_maps=None, test_no=8):
    # List of ports to bind to.
    local_ports = []

    # Remote ports that our NAT will show.
    remote_ports = []

    # Indicates whether replies need to be from a certain remote port.
    reply_ports = []

    # Stun socks (if used.)
    stun_socks = []

    # Patch NAT if it's not remote.
    if mode in [TCP_PUNCH_LAN, TCP_PUNCH_SELF]:
        our_nat = nat_info(OPEN_INTERNET, delta_info(NA_DELTA, 0))
        their_nat = nat_info(OPEN_INTERNET, delta_info(NA_DELTA, 0))

    # Convenience func.
    def save_mapping(bind_p, remote_p=None, reply_p=0):
        if remote_p is None:
            remote_p = bind_p

        local_ports.append(bind_p)
        remote_ports.append(remote_p)
        reply_ports.append(reply_p)

    # Set test_no based on recipients test no.
    # [[remote port, required reply port], ...]
    if their_maps is not None:
        test_no = len(their_maps)

    # Pretend we have a list of mappings from a peer
    # even if we don't -- used to simplify code.
    use_stun_port = nats_can_predict(our_nat, their_nat)
    use_range = nats_intersect_range(our_nat, their_nat, test_no)
    if their_maps is None:
        their_maps = []
        for i in range(0, test_no):
            # Default port for when there is a [port restrict, rand delta] NAT.
            if use_stun_port:
                test_no = 1
                their_maps.append([STUN_PORT, 0, 0])
                break
            else:
                their_maps.append([from_range(use_range), 0, 0])

    # Set initial mapping needed for delta type NATs.
    last_mapped = [None, None]
    if our_nat["delta"]["type"] in DELTA_N:
        # NOTE: if local < 1024 it may not be possible to reuse.
        r = stun_client.interface.route(stun_client.af)
        s = None
        for i in range(0, 5):
            try:
                # Reserve a sock for use.
                # Close it so it can be used.
                reserved_sock, high_port = await get_high_port_socket(r, sock_type=TCP)
                #reserved_sock.close()

                
                _, s, local_port, remote_port, _, _ = await stun_client.get_mapping(
                    proto=STREAM,
                    source_port=high_port
                )

                #socket.setsockopt(s, socket.SO_REUSEPORT)
                last_mapped = [local_port, remote_port]
                break
            except:
                log_exception()
                s = None
                continue

        if s is None:
            raise Exception("high port sock fail in init map punch.")
        stun_socks.append(s)

    # Make a list of tasks to build port predictions.
    # Use default ports for client if unknown or try use their ports.
    tasks = []
    results = []
    for i in range(0, test_no):
        # Predict our mappings.
        # Try to match our ports to any provided mappings.
        task = get_single_mapping(
            # Punching mode.
            mode,

            # Try match this mapping.
            their_maps[i],

            # A mapping fetch from STUN.
            # Only set depending on certain NATs.
            last_mapped,

            # Uses an range compatible with both NATs for mappings.
            # Otherwise uses a range compatible with our_nat.
            use_range,

            # Info on our NAT type and delta.
            our_nat,

            # Reference to do a STUN request to get a mapping.
            stun_client
        )

        """
        Concurrent NAT types can use asyncio.gather.
        Non-concurrent NATs need to be executed in order.
        The reason for this is that on non-concurrent NATs
        the allocations rely on using an incrementing sequence
        of source ports. The order these are allocated couldn't
        be guaranteed if asyncio.gather were to be used.
        """
        if our_nat["is_concurrent"]:
            tasks.append(task)
        else:
            result = await task
            results.append(result)

    # Run tasks concurrently.
    if len(tasks):
        results = await asyncio.gather(*tasks)

    # Store results in correct indexes.
    for result in results:
        info, _ = result
        local, remote, reply, sock = info
        save_mapping(local, remote, reply)
        if sock is not None:
            stun_socks.append(sock)

    # Invalid state -- caller handle this pl0x.
    if not(len(local_ports)):
        raise Exception("Unable to predict mappings")

    ###############################################################
    # Make a list of duplicate offsets to strip.
    assert(len(local_ports) == len(remote_ports))
    assert(len(remote_ports) == len(reply_ports))
    check_lists = [ local_ports, remote_ports, reply_ports ]
    dups = [ [], [], [] ]
    offending_offsets = []
    for j in range(0, len(check_lists)):
        check_list = check_lists[j]
        for i in range(0, len(check_list)):
            if check_list in [ local_ports, reply_ports ]:
                if not check_list[i]:
                    continue

            if check_list[i] in dups[j]:
                offending_offsets.append(i)
                continue

            dups[j].append(check_list[i])
            
    # Delete the offending elements.
    # Offsets are adjusted to account for the
    # changed / new size of the list after changes.
    offending_offsets = sorted(offending_offsets)
    offending_offsets = to_unique(offending_offsets)
    for i in range(0, len(offending_offsets)):
        offset = offending_offsets[i] - i
        for check_list in check_lists:
            del check_list[offset]

    # Local ports are relative to a given NIC IP
    # hence we also return the interface.
    return {
        "interface": stun_client.interface,
        "af": stun_client.af,
        "local": local_ports,
        "remote": remote_ports,
        "reply": reply_ports,
        "stun_socks": stun_socks,
        "default_maps": their_maps,
        "last_mapped": last_mapped
    }

"""
It's extremely important that all predicts are above a certain
value otherwise admin or root will be needed to bind to the port
"""
def nat_check_for_low_ports(map_info):
    for k in ["local"]:
        for p in map_info[k]:
            if p <= 1024:
                error = f"invalid low port found for mapping {p}"
                log(error)
                raise Exception(error)

async def nat_predict_main():
    from .stun_client import STUNClient
    from .interface import Interface, init_p2pd

    # Load default interface.
    i = await Interface().start_local()
    s = STUNClient(i)
    
    # Run delta test and profile it.
    s1 = timestamp(1)
    out = await delta_test(s)
    print(f"delta test time = {timestamp(1) - s1}")
    print(out)

    # 0.36

if __name__ == "__main__":
    async_test(nat_predict_main)


import socket
from struct import pack
import hmac
from hashlib import sha1
from .net import *
from .settings import *

# Config variables -------------------------------------

TURN_MAX_RETRANSMITS = 5
TURN_MAIN_REPLY_TIMEOUT = 5

# secs - Turn recommends 1 minute before expiry.
TURN_REFRESH_EXPIRY = 600 
TURN_MAX_DICT_LEN = 1000
TURN_MAX_RECV_PACKETS = 100

#########################################################
TURN_MAGIC_COOKIE = b"\x21\x12\xA4\x42"
TURN_MAGIC_XOR = b'\x00\x00\x21\x12\x21\x12\xa4\x42'
TURN_CHANNEL = b"\x40\x02\x00\x00"
TURN_PROTOCOL_TCP = b"\x06\x00\x00\x00"
TURN_RPOTOCOL_UDP = b"\x11\x00\x00\x00"
TURN_CHAN_RANGE = [16384, 32766]

# Protocol state machine.
# With convenient lookup values.
TURN_NOT_STARTED = 1
TURN_TRY_ALLOCATE = 2
TURN_ALLOCATE_FAILED = 3
TURN_TRY_REQ_TRANSPORT = 4
TURN_REQ_TRANSPORT_FAILED = 5
TURN_TRY_REFRESH = 6
TURN_REFRESH_DONE = 7
TURN_REFRESH_FAIL = 8
TURN_ERROR_STOPPED = 9

def _get_const_name(cls, val, type_: type) -> str:
    for attr_name in dir(cls):
        attr = getattr(cls, attr_name)
        if isinstance(attr, type_) and attr == val:
            return attr_name
    return ''

# rfc5389#page-36 error no section.
def turn_write_error_attr(error_no, error_msg):
    if error_no < 300 or error_no > 699:
        raise Exception("TURN error no must be between 300 and 699")

    error_msg_bytes = error_msg.encode("utf-8")
    if len(error_msg_bytes) > 763:
        raise Exception("TURN error message is too long.")

    # Reserved portion.
    # Used to pad to a 32 bit boundary.
    bound = "0" * 21

    # The error_no is between 300 and 699.
    # The class bits take the 3 or 6 from the error no and convert to binary.
    # It is padded with zeros to ensure its 3 bits.
    bound += (("0" * 3) + bin(int(str(error_no)[0]))[2:])[-3:]

    # No may be 8 bits.
    bound += (("0" * 8) + bin(error_no % 100)[2:])[-8:]

    # Convert binary string to bytes (4).
    hdr = bits_to_bytes(bound)
    return hdr + error_msg_bytes

# Extract a tuple of a remote IP + port from a TURN attribute.
def turn_peer_attr_to_tup(peer_data, txid, af):
    if af == socket.AF_INET:
        # IPv4 and port.
        addr = TurnIPAddress(do_xor=True, xor_extra=txid).unpack(peer_data, family=b"\x00\x01")
    else:
        # IPv6 - needs extra XOR part.
        addr = TurnIPAddress(
            do_xor=True,
            xor_extra=txid
        ).unpack(peer_data, family=b"\x00\x02")

    return (addr.ip, int(addr.port))

class TurnMessageMethod:
    Reversed            = b"\x00\x00" # RFC5389
    Binding             = b"\x00\x01"
    SharedSecret        = b"\x00\x02" # Reserved - RFC5389
    
    # https://tools.ietf.org/html/rfc5766#section-13
    # https://datatracker.ietf.org/doc/html/draft-rosenberg-midcom-turn-08#section-9.1
    Allocate            = b"\x00\x03"
    Refresh             = b"\x00\x04"
    Send                = b"\x00\x06" # 4 or 6? send was previously wrong
    SendResponse        = b"\x01\x04"
    DataIndication      = b"\x01\x15"
    Data                = b"\x00\x07"
    CreatePermission    = b"\x00\x08"
    ChannelBind         = b"\x00\x09"

    # https://tools.ietf.org/html/rfc6062#section-6.1
    Connect             = b"\x00\x0a" # RFC6062
    ConnectionBind      = b"\x00\x0b" # RFC6062
    ConnectionAttempt   = b"\x00\x0c" # RFC6062

    get = classmethod(lambda cls, val, type_=int: _get_const_name(cls, val, type_))

class TurnMessageCode:
    Request     = b"\x00\x00"
    Indication  = b"\x00\x10"
    SuccessResp = b"\x01\x00"
    ErrorResp   = b"\x01\x10"

    get = classmethod(lambda cls, val, type_=int: _get_const_name(cls, val, type_))

class TurnAttribute:
    Reserved            = b"\x00\x00" # RFC5389
    MappedAddress       = b"\x00\x01" # RFC5389
    ResponseAddress     = b"\x00\x02" # Reserved - RFC5389
    ChangeRequest       = b"\x00\x03" # Reserved - RFC5389
    SourceAddress       = b"\x00\x04" # Reserved - RFC5389
    ChangedAddress      = b"\x00\x05" # Reserved - RFC5389
    Username            = b"\x00\x06" # RFC5389
    Password            = b"\x00\x07" # Reserved - RFC5389
    MessageIntegrity    = b"\x00\x08" # RFC5389
    ErrorCode           = b"\x00\x09" # RFC5389
    UnknownAttribute    = b"\x00\x0A" # RFC5389
    ReflectedFrom       = b"\x00\x0B" # Reserved - RFC5389


    ChannelNumber       = b"\x00\x0C" # RFC5766
    Lifetime            = b"\x00\x0D" # RFC5766
    Bandwidth           = b"\x00\x10" # Reserved - RFC5766
    DestinationAddress  = b"\x00\x11"

    XorPeerAddress      = b"\x00\x12" # RFC5766
    Data                = b"\x00\x13" # RFC5766

    Realm               = b"\x00\x14" # RFC5389
    Nonce               = b"\x00\x15" # RFC5389
    
    XorRelayedAddress   = b"\x00\x16" # RFC5766
    EvenPort            = b"\x00\x18" # RFC5766
    RequestedTransport  = b"\x00\x19" # RFC5766
    RequestedAddress    = b"\x00\x17" # draft-ietf-behave-turn-ipv6
    DontFragment        = b"\x00\x1A" # RFC5766

    XorMappedAddress    = b"\x00\x20" # RFC5389

    TimerVal            = b"\x00\x21" # Reserved - RFC5766
    ReservationToken    = b"\x00\x22" # RFC5766

    ConnectionID        = b"\x00\x2A" # RFC6062

    XorMappedAddressX   = b"\x80\x20"
    Software            = b"\x80\x22" # RFC5389
    AlternateServer     = b"\x80\x23" # RFC5389
    Fingerprint         = b"\x80\x28" # RFC5389
    UnknownAddress2     = b"\x80\x2b"
    UnknownAddress3     = b"\x80\x2c"
    
    get = classmethod(lambda cls, val, type_=bytes: _get_const_name(cls, val, type_))

class TurnIPAddress:
    def __init__(self,  ip=None, port=None, family=b"\x00\x01", do_xor=True, xor_extra=b""):
        self.family = family # 1 for ipv4, 2 for ipv6
        self.af = None
        self.ip = ip # type: str
        self.ip_b = None
        self.port = port 
        self.do_xor = do_xor
        self.xor_extra = xor_extra
        if ip is not None:
            if family == b"\x00\x01":
                self.ip_b = socket.inet_pton(
                    socket.AF_INET,
                    self.ip
                )
            else:
                self.ip_b = socket.inet_pton(
                    socket.AF_INET6,
                    self.ip
                )

    def _xor_cookie(self, a):
        # Make virtual bytearray using pointers.
        b = bytearray().join([
            memoryview(TURN_MAGIC_XOR),
            memoryview(self.xor_extra)
        ])

        # Create a list of ints representing XOR cookie.
        r = bytearray(); b_len = len(b); a_len = len(a);
        for i in range(0, max(a_len, b_len) ):
            r.append(a[i % a_len] ^ b[i % b_len])

        # Return cookie as bytearray to avoid making a copy.
        return r

    def encode(self) -> bytes:
        # Avoid copying fields as much as possible.
        buf = bytearray().join([
            self.family,
            memoryview(pack('!H', self.port)),
            self.ip_b
        ])

        # XOR IPAddress required for IPv6.
        if self.do_xor:
            buf = self._xor_cookie(buf)

        return buf

    def decode(self, data: bytes):
        if self.do_xor:
            data = self._xor_cookie(data)
        data = memoryview(data)
        self.family = data[0:2]
        self.port = b_to_i(data[2:4], 'big')
        if self.family == b"\x00\x01":
            self.ip_b = data[4:8]
            self.af = socket.AF_INET
        else:
            self.ip_b = data[4:20]
            self.af = socket.AF_INET6

        self.ip = socket.inet_ntop(self.af, self.ip_b)

    def pack(self, ip: str, port: int, family: int) -> bytes:
        inst = TurnIPAddress(ip=ip, port=port, do_xor=self.do_xor, family=family, xor_extra=self.xor_extra)
        return inst.encode()

    def unpack(self, data: bytes, family: int):
        inst = TurnIPAddress(do_xor=self.do_xor, family=family, xor_extra=self.xor_extra)
        inst.decode(data)
        return inst

    def __str__(self):
        return '{}:{}'.format(self.ip, self.port)

class TurnMessage:
    def __init__(self, msg_type=None, msg_code=TurnMessageCode.Request):
        self.msg_code = msg_code
        self.msg_type = msg_type # type: int
        self.msg_len = 0 # type: int
        self.magic_cookie = TURN_MAGIC_COOKIE
        self.txn_id = memoryview(rand_b(12)) # type: bytes
        self.msg = bytearray()
        self.attr_cursor = 0 # type: int

    def reset_attr(self):
        self.msg_len = 0
        self.msg = bytearray()

    def write_attr(self, attr: bytes, *data, fmt: str = None):
        # process data -> bytes
        if fmt:
            data = pack(fmt, *data)
        else:
            data = data[0]
            if isinstance(data, TurnIPAddress):
                data = data.encode()

        # Rule of 4:
        # https://tools.ietf.org/html/rfc5766#section-14
        padding = b""
        if len(data) % 4 != 0:
            padding = b'\x00' * (4 - len(data) % 4)

        buf = bytearray().join([
            memoryview(attr),
            memoryview(pack("!H", len(data))),
            memoryview(data),
            memoryview(padding)
        ])

        self.msg_len += len(buf)
        self.msg.extend(buf)

    def write_credential(self, username: str, realm: str, nonce: bytes = b''):
        self.write_attr(TurnAttribute.Username, username)
        self.write_attr(TurnAttribute.Realm, realm)
        self.write_attr(TurnAttribute.Nonce, nonce)

    def _hmac(self, key: bytes, msg: bytes) -> bytes:
        hashed = hmac.new(key, msg, sha1)
        return hashed.digest()
    
    def write_hmac(self, key: bytes):
        self.msg_len += 24
        msg_hmac = self.encode()
        self.msg_len -= 24
        self.write_attr(TurnAttribute.MessageIntegrity, self._hmac(key, msg_hmac))

    def eof(self) -> bool:
        return self.attr_cursor >= self.msg_len - 1

    def read_attr(self) -> tuple:
        # Process serialized attribute chunk using pointers.
        msg = memoryview(self.msg)
        msg_len = len(msg)
        m_attr = m_len = m_data = None
        if msg_len and self.attr_cursor + 3 <= msg_len - 1:
            # Unpack first two fields of an attribute.
            attr_hdr = msg[self.attr_cursor:self.attr_cursor + 4]
            m_attr = attr_hdr[0:2]
            m_len = b_to_i(attr_hdr[2:4], 'big')

            # Avoid overflows for attribute data.
            self.attr_cursor += 4
            if m_len:
                if self.attr_cursor + (m_len - 1) <= msg_len - 1:
                    # Get attribute data.
                    m_data = msg[self.attr_cursor:self.attr_cursor + m_len]

                    # Rule of block 4: 
                    # https://tools.ietf.org/html/rfc5766#section-14
                    attr_pad = 0
                    if m_len % 4 != 0:
                        attr_pad = 4 - m_len % 4

                    # Increase attribute pointer.
                    self.attr_cursor += m_len + attr_pad
                else:
                    raise Exception("TURN attribute len invalid.")

        # Return results.
        return m_attr, m_len, m_data 


    def encode(self) -> bytes:
        return bytearray().join([
            memoryview(b_and(b_or(self.msg_type, self.msg_code), b"\x3F\xFF")),
            memoryview(pack("!H", self.msg_len)),
            memoryview(self.magic_cookie),
            memoryview(self.txn_id),
            memoryview(self.msg)
        ])

    def decode(self, msg: memoryview) -> memoryview:
        # Unpack data from buffer using memory views.
        msg_len = len(msg)
        self.attr_cursor = 0
        if msg_len >= 20:
            # Unpack message fields.
            self.msg_type = msg[0:2]
            self.msg_len = b_to_i(msg[2:4], 'big')
            self.magic_cookie = msg[4:8]
            self.txn_id = msg[8:20]

            # Make sure message len accurately reflects size.
            if self.msg_len:
                if 20 + (self.msg_len - 1) <= msg_len - 1:
                    self.msg = msg[20:20 + self.msg_len]

                    # ret data left in buffer, usually NULL
                    return msg[20 + self.msg_len:]
                else:
                    raise Exception("Invalid length for TURN msg.")

    @classmethod
    def pack(cls, msg_type: bytes, msg: memoryview) -> memoryview:
        inst = cls(msg_type=msg_type, msg=msg, msg_len=len(msg))
        return inst.encode()
    
    @classmethod
    def unpack(cls, msg: memoryview) -> tuple:
        inst = cls()
        buf = inst.decode(msg)
        return inst, buf

# Writes an IP, port tuple as an attribute to a message.
def turn_write_peer_addr(msg, client_tup, attr_type=TurnAttribute.XorPeerAddress):
    ip_obj = ip_f(client_tup[0])
    af = v_to_af(ip_obj.version)
    peer_host, peer_port = client_tup[:2]
    af = b"\x00\x01" if af == socket.AF_INET else b"\x00\x02"
    xor_extra = b"" if af == b"\x00\x01" else msg.txn_id
    msg.write_attr(attr_type, TurnIPAddress(do_xor=True, ip=peer_host, port=peer_port, family=af, xor_extra=xor_extra))

def turn_vars_to_server(var_list, af):
    return {
        "host": var_list[0],
        "port": var_list[1],
        "user": var_list[2],
        "pass": var_list[3],
        "realm": var_list[4]
    }
    
def find_turn_server(turn_server, turn_servers, af=None):
    for needle in turn_servers:
        # Not the same server host or IP.
        if needle["host"] != turn_server["host"]:
            continue

        # Not the same port.
        if needle["port"] != turn_server["port"]:
            continue

        # Not the same user.
        if needle["user"] != turn_server["user"]:
            continue

        # Not the same pass.
        if needle["pass"] != turn_server["pass"]:
            continue

        # Not the same realm.
        if needle["realm"] != turn_server["realm"]:
            continue

        # Not the same AF.
        if af is not None:
            if af not in needle["afs"]:
                continue

        return True

    return False